"""Cache for commonly shared data that is intended to be immutable."""

from __future__ import annotations

import collections.abc as c
import typing as t

from .config import (
    CommonConfig,
)

TValue = t.TypeVar('TValue')


class CommonCache:
    """Common cache."""

    def __init__(self, args: CommonConfig) -> None:
        self.args = args

    def get(self, key: str, factory: c.Callable[[], TValue]) -> TValue:
        """Return the value from the cache identified by the given key, using the specified factory method if it is not found."""
        if key not in self.args.cache:
            self.args.cache[key] = factory()

        return self.args.cache[key]

    def get_with_args(self, key: str, factory: c.Callable[[CommonConfig], TValue]) -> TValue:
        """Return the value from the cache identified by the given key, using the specified factory method (which accepts args) if it is not found."""
        if key not in self.args.cache:
            self.args.cache[key] = factory(self.args)

        return self.args.cache[key]



# This file is part of Ansible
# -*- coding: utf-8 -*-
#
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

# for testing
import unittest
from unittest.mock import MagicMock

from ansible.module_utils.facts.sysctl import get_sysctl


# `sysctl hw` on an openbsd machine
OPENBSD_SYSCTL_HW = """
hw.machine=amd64
hw.model=AMD EPYC Processor (with IBPB)
hw.ncpu=1
hw.byteorder=1234
hw.pagesize=4096
hw.disknames=cd0:,sd0:9e1bd96cb20ab429,fd0:
hw.diskcount=3
hw.sensors.viomb0.raw0=0 (desired)
hw.sensors.viomb0.raw1=0 (current)
hw.cpuspeed=3394
hw.vendor=QEMU
hw.product=Standard PC (i440FX + PIIX, 1996)
hw.version=pc-i440fx-5.1
hw.uuid=5833415a-eefc-964f-a306-fa434d44d117
hw.physmem=1056804864
hw.usermem=1056792576
hw.ncpufound=1
hw.allowpowerdown=1
hw.smt=0
hw.ncpuonline=1
"""

# partial output of `sysctl kern` on an openbsd machine
# for testing multiline parsing
OPENBSD_SYSCTL_KERN_PARTIAL = """
kern.ostype=OpenBSD
kern.osrelease=6.7
kern.osrevision=202005
kern.version=OpenBSD 6.7 (GENERIC) #179: Thu May  7 11:02:37 MDT 2020
    deraadt@amd64.openbsd.org:/usr/src/sys/arch/amd64/compile/GENERIC

kern.maxvnodes=12447
kern.maxproc=1310
kern.maxfiles=7030
kern.argmax=524288
kern.securelevel=1
kern.hostname=openbsd67.vm.home.elrod.me
kern.hostid=0
kern.clockrate=tick = 10000, tickadj = 40, hz = 100, profhz = 100, stathz = 100
kern.posix1version=200809
"""

# partial output of `sysctl vm` on Linux. The output has tabs in it and Linux
# sysctl has spaces around the =
LINUX_SYSCTL_VM_PARTIAL = """
vm.dirty_background_ratio = 10
vm.dirty_bytes = 0
vm.dirty_expire_centisecs = 3000
vm.dirty_ratio = 20
vm.dirty_writeback_centisecs = 500
vm.dirtytime_expire_seconds = 43200
vm.extfrag_threshold = 500
vm.hugetlb_shm_group = 0
vm.laptop_mode = 0
vm.legacy_va_layout = 0
vm.lowmem_reserve_ratio = 256	256	32	0
vm.max_map_count = 65530
vm.min_free_kbytes = 22914
vm.min_slab_ratio = 5
"""

# partial output of `sysctl vm` on macOS. The output is colon-separated.
MACOS_SYSCTL_VM_PARTIAL = """
vm.loadavg: { 1.28 1.18 1.13 }
vm.swapusage: total = 2048.00M  used = 1017.50M  free = 1030.50M  (encrypted)
vm.cs_force_kill: 0
vm.cs_force_hard: 0
vm.cs_debug: 0
vm.cs_debug_fail_on_unsigned_code: 0
vm.cs_debug_unsigned_exec_failures: 0
vm.cs_debug_unsigned_mmap_failures: 0
vm.cs_all_vnodes: 0
vm.cs_system_enforcement: 1
vm.cs_process_enforcement: 0
vm.cs_enforcement_panic: 0
vm.cs_library_validation: 0
vm.global_user_wire_limit: 3006477107
"""

# Invalid/bad output
BAD_SYSCTL = """
this.output.is.invalid
it.has.no.equals.sign.or.colon
so.it.should.fail.to.parse
"""

# Mixed good/bad output
GOOD_BAD_SYSCTL = """
bad.output.here
hw.smt=0
and.bad.output.here
"""


class TestSysctlParsingInFacts(unittest.TestCase):

    def test_get_sysctl_missing_binary(self):
        module = MagicMock()
        module.get_bin_path.return_value = '/usr/sbin/sysctl'
        module.run_command.side_effect = ValueError
        self.assertRaises(ValueError, get_sysctl, module, ['vm'])

    def test_get_sysctl_nonzero_rc(self):
        module = MagicMock()
        module.get_bin_path.return_value = '/usr/sbin/sysctl'
        module.run_command.return_value = (1, '', '')
        sysctl = get_sysctl(module, ['hw'])
        self.assertEqual(sysctl, {})

    def test_get_sysctl_command_error(self):
        module = MagicMock()
        module.get_bin_path.return_value = '/usr/sbin/sysctl'
        module.reset_mock()
        module.run_command.side_effect = OSError('foo')
        sysctl = get_sysctl(module, ['hw'])
        for call in module.error_as_warning.call_args_list:
            self.assertIn('Unable to read sysctl', call[0][0])
        self.assertEqual(sysctl, {})

    def test_get_sysctl_all_invalid_output(self):
        module = MagicMock()
        module.get_bin_path.return_value = '/sbin/sysctl'
        module.run_command.return_value = (0, BAD_SYSCTL, '')
        sysctl = get_sysctl(module, ['hw'])
        module.run_command.assert_called_once_with(['/sbin/sysctl', 'hw'])
        lines = [l for l in BAD_SYSCTL.splitlines() if l]
        for call in module.error_as_warning.call_args_list:
            self.assertIn('Unable to split sysctl line', call[0][0])
        self.assertEqual(module.error_as_warning.call_count, len(lines))
        self.assertEqual(sysctl, {})

    def test_get_sysctl_mixed_invalid_output(self):
        module = MagicMock()
        module.get_bin_path.return_value = '/sbin/sysctl'
        module.run_command.return_value = (0, GOOD_BAD_SYSCTL, '')
        sysctl = get_sysctl(module, ['hw'])
        module.run_command.assert_called_once_with(['/sbin/sysctl', 'hw'])
        bad_lines = ['bad.output.here', 'and.bad.output.here']
        for call in module.error_as_warning.call_args_list:
            self.assertIn('Unable to split sysctl line', call[0][0])
        self.assertEqual(module.error_as_warning.call_count, 2)
        self.assertEqual(sysctl, {'hw.smt': '0'})

    def test_get_sysctl_openbsd_hw(self):
        expected_lines = [l for l in OPENBSD_SYSCTL_HW.splitlines() if l]
        module = MagicMock()
        module.get_bin_path.return_value = '/sbin/sysctl'
        module.run_command.return_value = (0, OPENBSD_SYSCTL_HW, '')
        sysctl = get_sysctl(module, ['hw'])
        module.run_command.assert_called_once_with(['/sbin/sysctl', 'hw'])
        self.assertEqual(len(sysctl), len(expected_lines))
        self.assertEqual(sysctl['hw.machine'], 'amd64')  # first line
        self.assertEqual(sysctl['hw.smt'], '0')  # random line
        self.assertEqual(sysctl['hw.ncpuonline'], '1')  # last line
        # weird chars in value
        self.assertEqual(
            sysctl['hw.disknames'],
            'cd0:,sd0:9e1bd96cb20ab429,fd0:')
        # more symbols/spaces in value
        self.assertEqual(
            sysctl['hw.product'],
            'Standard PC (i440FX + PIIX, 1996)')

    def test_get_sysctl_openbsd_kern(self):
        module = MagicMock()
        module.get_bin_path.return_value = '/sbin/sysctl'
        module.run_command.return_value = (0, OPENBSD_SYSCTL_KERN_PARTIAL, '')
        sysctl = get_sysctl(module, ['kern'])
        module.run_command.assert_called_once_with(['/sbin/sysctl', 'kern'])
        self.assertEqual(
            len(sysctl),
            len(
                [l for l
                 in OPENBSD_SYSCTL_KERN_PARTIAL.splitlines()
                 if l.startswith('kern')]))
        self.assertEqual(sysctl['kern.ostype'], 'OpenBSD')  # first line
        self.assertEqual(sysctl['kern.maxproc'], '1310')  # random line
        self.assertEqual(sysctl['kern.posix1version'], '200809')  # last line
        # multiline
        self.assertEqual(
            sysctl['kern.version'],
            'OpenBSD 6.7 (GENERIC) #179: Thu May  7 11:02:37 MDT 2020\n    '
            'deraadt@amd64.openbsd.org:/usr/src/sys/arch/amd64/compile/GENERIC')
        # more symbols/spaces in value
        self.assertEqual(
            sysctl['kern.clockrate'],
            'tick = 10000, tickadj = 40, hz = 100, profhz = 100, stathz = 100')

    def test_get_sysctl_linux_vm(self):
        module = MagicMock()
        module.get_bin_path.return_value = '/usr/sbin/sysctl'
        module.run_command.return_value = (0, LINUX_SYSCTL_VM_PARTIAL, '')
        sysctl = get_sysctl(module, ['vm'])
        module.run_command.assert_called_once_with(['/usr/sbin/sysctl', 'vm'])
        self.assertEqual(
            len(sysctl),
            len([l for l in LINUX_SYSCTL_VM_PARTIAL.splitlines() if l]))
        self.assertEqual(sysctl['vm.dirty_background_ratio'], '10')
        self.assertEqual(sysctl['vm.laptop_mode'], '0')
        self.assertEqual(sysctl['vm.min_slab_ratio'], '5')
        # tabs
        self.assertEqual(sysctl['vm.lowmem_reserve_ratio'], '256\t256\t32\t0')

    def test_get_sysctl_macos_vm(self):
        module = MagicMock()
        module.get_bin_path.return_value = '/usr/sbin/sysctl'
        module.run_command.return_value = (0, MACOS_SYSCTL_VM_PARTIAL, '')
        sysctl = get_sysctl(module, ['vm'])
        module.run_command.assert_called_once_with(['/usr/sbin/sysctl', 'vm'])
        self.assertEqual(
            len(sysctl),
            len([l for l in MACOS_SYSCTL_VM_PARTIAL.splitlines() if l]))
        self.assertEqual(sysctl['vm.loadavg'], '{ 1.28 1.18 1.13 }')
        self.assertEqual(
            sysctl['vm.swapusage'],
            'total = 2048.00M  used = 1017.50M  free = 1030.50M  (encrypted)')



from __future__ import annotations

import abc
import copy
import typing as t

from yaml import Node, ScalarNode
from yaml.constructor import SafeConstructor
from yaml.resolver import BaseResolver

from ansible import constants as C
from ansible.module_utils.common.text.converters import to_text
from ansible.module_utils._internal._datatag import AnsibleTagHelper, AnsibleDatatagBase
from ansible._internal._datatag._tags import Origin, TrustedAsTemplate
from ansible.parsing.vault import EncryptedString
from ansible.utils.display import Display

from ._errors import AnsibleConstructorError

display = Display()

_TRUSTED_AS_TEMPLATE: t.Final[TrustedAsTemplate] = TrustedAsTemplate()


class _BaseConstructor(SafeConstructor, metaclass=abc.ABCMeta):
    """Base class for Ansible YAML constructors."""

    @classmethod
    @abc.abstractmethod
    def _register_constructors(cls) -> None:
        """Method used to register constructors to derived types during class initialization."""

    def __init_subclass__(cls, **kwargs) -> None:
        """Initialization for derived types."""
        cls._register_constructors()


class AnsibleInstrumentedConstructor(_BaseConstructor):
    """Ansible constructor which supports Ansible custom behavior such as `Origin` tagging, but no Ansible-specific YAML tags."""

    name: t.Any  # provided by the YAML parser, which retrieves it from the stream

    def __init__(self, origin: Origin, trusted_as_template: bool) -> None:
        if not origin.line_num:
            origin = origin.replace(line_num=1)

        self._origin = origin
        self._trusted_as_template = trusted_as_template
        self._duplicate_key_mode = C.config.get_config_value('DUPLICATE_YAML_DICT_KEY')

        super().__init__()

    @property
    def trusted_as_template(self) -> bool:
        return self._trusted_as_template

    def construct_yaml_map(self, node):
        data = self._node_position_info(node).tag({})  # always an ordered dictionary on py3.7+
        yield data
        value = self.construct_mapping(node)
        data.update(value)

    def construct_mapping(self, node, deep=False):
        # Delegate to built-in implementation to construct the mapping.
        # This is done before checking for duplicates to leverage existing error checking on the input node.
        mapping = super().construct_mapping(node, deep)
        keys = set()

        # Now that the node is known to be a valid mapping, handle any duplicate keys.
        for key_node, _value_node in node.value:
            if (key := self.construct_object(key_node, deep=deep)) in keys:
                msg = f'Found duplicate mapping key {key!r}.'

                if self._duplicate_key_mode == 'error':
                    raise AnsibleConstructorError(problem=msg, problem_mark=key_node.start_mark)

                if self._duplicate_key_mode == 'warn':
                    display.warning(msg=msg, obj=key, help_text='Using last defined value only.')

            keys.add(key)

        return mapping

    def construct_yaml_int(self, node):
        value = super().construct_yaml_int(node)
        return self._node_position_info(node).tag(value)

    def construct_yaml_float(self, node):
        value = super().construct_yaml_float(node)
        return self._node_position_info(node).tag(value)

    def construct_yaml_timestamp(self, node):
        value = super().construct_yaml_timestamp(node)
        return self._node_position_info(node).tag(value)

    def construct_yaml_omap(self, node):
        origin = self._node_position_info(node)
        display.deprecated(
            msg='Use of the YAML `!!omap` tag is deprecated.',
            version='2.23',
            obj=origin,
            help_text='Use a standard mapping instead, as key order is always preserved.',
        )
        items = list(super().construct_yaml_omap(node))[0]
        items = [origin.tag(item) for item in items]
        yield origin.tag(items)

    def construct_yaml_pairs(self, node):
        origin = self._node_position_info(node)
        display.deprecated(
            msg='Use of the YAML `!!pairs` tag is deprecated.',
            version='2.23',
            obj=origin,
            help_text='Use a standard mapping instead.',
        )
        items = list(super().construct_yaml_pairs(node))[0]
        items = [origin.tag(item) for item in items]
        yield origin.tag(items)

    def construct_yaml_str(self, node: ScalarNode) -> str:
        # Override the default string handling function
        # to always return unicode objects
        # DTFIX-FUTURE: is this to_text conversion still necessary under Py3?
        value = to_text(self.construct_scalar(node))

        tags: list[AnsibleDatatagBase] = [self._node_position_info(node)]

        if self.trusted_as_template:
            # NB: since we're not context aware, this will happily add trust to dictionary keys; this is actually necessary for
            #  certain backward compat scenarios, though might be accomplished in other ways if we wanted to avoid trusting keys in
            #  the general scenario
            tags.append(_TRUSTED_AS_TEMPLATE)

        return AnsibleTagHelper.tag(value, tags)

    def construct_yaml_binary(self, node):
        value = super().construct_yaml_binary(node)

        return AnsibleTagHelper.tag(value, self._node_position_info(node))

    def construct_yaml_set(self, node):
        data = AnsibleTagHelper.tag(set(), self._node_position_info(node))
        yield data
        value = self.construct_mapping(node)
        data.update(value)

    def construct_yaml_seq(self, node):
        data = self._node_position_info(node).tag([])
        yield data
        data.extend(self.construct_sequence(node))

    def _resolve_and_construct_object(self, node):
        # use a copied node to avoid mutating existing node and tripping the recursion check in construct_object
        copied_node = copy.copy(node)
        # repeat implicit resolution process to determine the proper tag for the value in the unsafe node
        copied_node.tag = t.cast(BaseResolver, self).resolve(type(node), node.value, (True, False))

        # re-entrant call using the correct tag
        # non-deferred construction of hierarchical nodes so the result is a fully realized object, and so our stateful unsafe propagation behavior works
        return self.construct_object(copied_node, deep=True)

    def _node_position_info(self, node) -> Origin:
        # the line number where the previous token has ended (plus empty lines)
        # Add one so that the first line is line 1 rather than line 0
        return self._origin.replace(line_num=node.start_mark.line + self._origin.line_num, col_num=node.start_mark.column + 1)

    @classmethod
    def _register_constructors(cls) -> None:
        constructors: dict[str, t.Callable] = {
            'tag:yaml.org,2002:binary': cls.construct_yaml_binary,
            'tag:yaml.org,2002:float': cls.construct_yaml_float,
            'tag:yaml.org,2002:int': cls.construct_yaml_int,
            'tag:yaml.org,2002:map': cls.construct_yaml_map,
            'tag:yaml.org,2002:omap': cls.construct_yaml_omap,
            'tag:yaml.org,2002:pairs': cls.construct_yaml_pairs,
            'tag:yaml.org,2002:python/dict': cls.construct_yaml_map,
            'tag:yaml.org,2002:python/unicode': cls.construct_yaml_str,
            'tag:yaml.org,2002:seq': cls.construct_yaml_seq,
            'tag:yaml.org,2002:set': cls.construct_yaml_set,
            'tag:yaml.org,2002:str': cls.construct_yaml_str,
            'tag:yaml.org,2002:timestamp': cls.construct_yaml_timestamp,
        }

        for tag, constructor in constructors.items():
            cls.add_constructor(tag, constructor)


class AnsibleConstructor(AnsibleInstrumentedConstructor):
    """Ansible constructor which supports Ansible custom behavior such as `Origin` tagging, as well as Ansible-specific YAML tags."""

    def __init__(self, origin: Origin, trusted_as_template: bool) -> None:
        self._unsafe_depth = 0  # volatile state var used during recursive construction of a value tagged unsafe

        super().__init__(origin=origin, trusted_as_template=trusted_as_template)

    @property
    def trusted_as_template(self) -> bool:
        return self._trusted_as_template and not self._unsafe_depth

    def construct_yaml_unsafe(self, node):
        self._unsafe_depth += 1

        try:
            return self._resolve_and_construct_object(node)
        finally:
            self._unsafe_depth -= 1

    def construct_yaml_vault(self, node: Node) -> EncryptedString:
        ciphertext = self._resolve_and_construct_object(node)

        if not isinstance(ciphertext, str):
            raise AnsibleConstructorError(problem=f"the {node.tag!r} tag requires a string value", problem_mark=node.start_mark)

        encrypted_string = AnsibleTagHelper.tag_copy(ciphertext, EncryptedString(ciphertext=AnsibleTagHelper.untag(ciphertext)))

        return encrypted_string

    def construct_yaml_vault_encrypted(self, node: Node) -> EncryptedString:
        origin = self._node_position_info(node)
        display.deprecated(
            msg='Use of the YAML `!vault-encrypted` tag is deprecated.',
            version='2.23',
            obj=origin,
            help_text='Use the `!vault` tag instead.',
        )

        return self.construct_yaml_vault(node)

    @classmethod
    def _register_constructors(cls) -> None:
        super()._register_constructors()

        constructors: dict[str, t.Callable] = {
            '!unsafe': cls.construct_yaml_unsafe,
            '!vault': cls.construct_yaml_vault,
            '!vault-encrypted': cls.construct_yaml_vault_encrypted,
        }

        for tag, constructor in constructors.items():
            cls.add_constructor(tag, constructor)



from __future__ import annotations

import typing as t

from . import _messages


class HasPluginInfo(t.Protocol):
    """Protocol to type-annotate and expose PluginLoader-set values."""

    @property
    def ansible_name(self) -> str | None:
        """Fully resolved plugin name."""

    @property
    def plugin_type(self) -> str:
        """Plugin type name."""


def get_plugin_info(value: HasPluginInfo) -> _messages.PluginInfo:
    """Utility method that returns a `PluginInfo` from an object implementing the `HasPluginInfo` protocol."""
    return _messages.PluginInfo(
        resolved_name=value.ansible_name,
        type=normalize_plugin_type(value.plugin_type),
    )


def normalize_plugin_type(value: str) -> _messages.PluginType | None:
    """Normalize value and return it as a PluginType, or None if the value does match any known plugin type."""
    value = value.lower()

    if value == 'modules':
        value = 'module'

    try:
        return _messages.PluginType(value)
    except ValueError:
        return None



# Copyright: (c) 2012, Michael DeHaan <michael.dehaan@gmail.com>
# Copyright: (c) 2012-17, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = """
    name: template
    author: Michael DeHaan
    version_added: "0.9"
    short_description: retrieve contents of file after templating with Jinja2
    description:
      - Returns a list of strings; for each template in the list of templates you pass in, returns a string containing the results of processing that template.
    options:
      _terms:
        description: list of files to template
      convert_data:
        type: bool
        description:
            - Whether to convert YAML into data. If V(False), strings that are YAML will be left untouched.
            - Mutually exclusive with the O(jinja2_native) option.
        default: true
        deprecated:
          why: This option is no longer used in the Ansible Core code base.
          version: "2.23"
          alternatives: Jinja2 native mode is now the default and only option, which is mutually exclusive with this option.
      variable_start_string:
        description: The string marking the beginning of a print statement.
        default: '{{'
        version_added: '2.8'
        type: str
      variable_end_string:
        description: The string marking the end of a print statement.
        default: '}}'
        version_added: '2.8'
        type: str
      jinja2_native:
        description:
            - Controls whether to use Jinja2 native types.
            - It is off by default even if global O(jinja2_native) is V(True).
            - Has no effect if global O(jinja2_native) is V(False).
            - This offers more flexibility than the template module which does not use Jinja2 native types at all.
        default: True
        version_added: '2.11'
        type: bool
        deprecated:
          why: This option is no longer used in the Ansible Core code base.
          version: "2.23"
          alternatives: Jinja2 native mode is now the default and only option.
      template_vars:
        description: A dictionary, the keys become additional variables available for templating.
        default: {}
        version_added: '2.3'
        type: dict
      comment_start_string:
        description: The string marking the beginning of a comment statement.
        version_added: '2.12'
        type: str
        default: '{#'
      comment_end_string:
        description: The string marking the end of a comment statement.
        version_added: '2.12'
        type: str
        default: '#}'
      trim_blocks:
        description:
        - Determine when newlines should be removed from blocks.
        - When set to V(yes) the first newline after a block is removed (block, not variable tag!).
        type: bool
        default: yes
        version_added: '2.19'
    seealso:
      - ref: playbook_task_paths
        description: Search paths used for relative templates.
"""

EXAMPLES = """
- name: show templating results
  ansible.builtin.debug:
    msg: "{{ lookup('ansible.builtin.template', './some_template.j2') }}"

- name: show templating results with different variable start and end string
  ansible.builtin.debug:
    msg: "{{ lookup('ansible.builtin.template', './some_template.j2', variable_start_string='[%', variable_end_string='%]') }}"

- name: show templating results with different comment start and end string
  ansible.builtin.debug:
    msg: "{{ lookup('ansible.builtin.template', './some_template.j2', comment_start_string='[#', comment_end_string='#]') }}"

- name: show templating results with trim_blocks
  ansible.builtin.debug:
    msg: "{{ lookup('ansible.builtin.template', './some_template.j2', trim_blocks=True) }}"

"""

RETURN = """
_raw:
   description: file(s) content after templating
   type: list
   elements: raw
"""

from copy import deepcopy
import os

from ansible.errors import AnsibleError
from ansible.plugins.lookup import LookupBase
from ansible.template import trust_as_template
from ansible._internal._templating import _template_vars
from ansible._internal._templating._engine import TemplateOptions, TemplateOverrides
from ansible.utils.display import Display


display = Display()


class LookupModule(LookupBase):

    def run(self, terms, variables=None, **kwargs):

        ret = []

        self.set_options(var_options=variables, direct=kwargs)

        # capture options
        lookup_template_vars = self.get_option('template_vars')
        variable_start_string = self.get_option('variable_start_string')
        variable_end_string = self.get_option('variable_end_string')
        comment_start_string = self.get_option('comment_start_string')
        comment_end_string = self.get_option('comment_end_string')
        trim_blocks = self.get_option('trim_blocks')

        templar = self._templar

        for term in terms:
            display.debug("File lookup term: %s" % term)

            lookupfile = self.find_file_in_search_path(variables, 'templates', term)
            display.vvvv("File lookup using %s as file" % lookupfile)
            if lookupfile:
                template_data = trust_as_template(self._loader.get_text_file_contents(lookupfile))

                # set jinja2 internal search path for includes
                searchpath = variables.get('ansible_search_path', [])
                if searchpath:
                    # our search paths aren't actually the proper ones for jinja includes.
                    # We want to search into the 'templates' subdir of each search path in
                    # addition to our original search paths.
                    newsearchpath = []
                    for p in searchpath:
                        newsearchpath.append(os.path.join(p, 'templates'))
                        newsearchpath.append(p)
                    searchpath = newsearchpath
                searchpath.insert(0, os.path.dirname(lookupfile))

                # The template will have access to all existing variables,
                # plus some added by ansible (e.g., template_{path,mtime}),
                # plus anything passed to the lookup with the template_vars=
                # argument.
                # FIXME: why isn't this a chainmap with a sacrificial bottom layer?
                vars = deepcopy(variables)
                vars.update(_template_vars.generate_ansible_template_vars(
                    path=term,
                    fullpath=lookupfile,
                    include_ansible_managed='ansible_managed' not in vars,  # do not clobber ansible_managed when set by the user
                ))
                vars.update(lookup_template_vars)

                overrides = dict(
                    variable_start_string=variable_start_string,
                    variable_end_string=variable_end_string,
                    comment_start_string=comment_start_string,
                    comment_end_string=comment_end_string,
                    trim_blocks=trim_blocks,
                )

                data_templar = templar.copy_with_new_env(available_variables=vars, searchpath=searchpath)
                # use the internal template API to avoid forced top-level finalization behavior imposed by the public API
                res = data_templar._engine.template(template_data, options=TemplateOptions(
                    escape_backslashes=False,
                    overrides=TemplateOverrides.from_kwargs(overrides),
                ))

                ret.append(res)
            else:
                raise AnsibleError("the template file %s could not be found for the lookup" % term)

        return ret



from __future__ import annotations

import collections.abc as c
import typing as t

import pytest

from ansible.module_utils._internal._datatag import (
    AnsibleSerializable,
    AnsibleTaggedObject,
)

# temporarily here to ensure this type is always "seen" in the known serializable types list
from ansible._internal._templating._jinja_bits import _JinjaConstTemplate  # pylint: disable=unused-import
from ansible._internal._templating._lazy_containers import _AnsibleLazyTemplateMixin, _AnsibleLazyAccessTuple

from ...module_utils.datatag.test_datatag import (
    ContainerTestCase,
    create_container_test_parameters,
    TestDatatagTarget as _TestDatatagTarget, Later, ParamDesc
)


class TestDatatagTemplar(_TestDatatagTarget):
    later = t.cast(t.Self, Later(locals(), _TestDatatagTarget))

    lazy_serializable_types: t.Annotated[
        list[type[c.Collection]], ParamDesc(["lazy_type"])
    ] = list(
        t.cast(type[c.Collection], known_type) for known_type in AnsibleSerializable._known_type_map.values()
        if issubclass(known_type, _AnsibleLazyTemplateMixin)
    )

    serializable_instances: t.Annotated[list[object], ParamDesc(["non_lazy_value"])]

    taggable_container_instances: t.Annotated[list[c.Collection], ParamDesc(["non_lazy_value"])] = _TestDatatagTarget.taggable_container_instances
    taggable_instances: t.Annotated[list[object], ParamDesc(["non_lazy_value"])] = t.cast(list[object], taggable_container_instances)

    @classmethod
    def post_init(cls, **kwargs):
        cls.serializable_types = t.cast(list[type[AnsibleSerializable]], cls.lazy_serializable_types)
        cls.serializable_instances = [(obj, ) for obj in cls.taggable_container_instances]

    @classmethod
    def container_test_parameters(cls, test_case: ContainerTestCase) -> tuple[t.Any, t.Optional[type], type]:
        """
        Return container test parameters for the given test case.
        Called during each test run to create the test value on-demand.
        """
        # pylint: disable=unidiomatic-typecheck
        candidates = [instance for instance in cls.taggable_container_instances if type(instance) is test_case.type_under_test.__mro__[2]]

        assert len(candidates) == 1

        value = candidates[0]
        value = _AnsibleLazyTemplateMixin._try_create(value)

        return create_container_test_parameters(test_case, value)

    @classmethod
    def container_test_cases(cls) -> t.Annotated[list[tuple[t.Any, type]], ParamDesc(["non_lazy_value", "type_under_test"])]:  # type: ignore[override]
        # for each lazy_serializable type, find exactly one matching taggable container instance
        # create the lazy type from the instance
        out_values = []
        for type_under_test in cls.lazy_serializable_types:
            candidates = [instance for instance in cls.taggable_container_instances
                          if type(instance) is type_under_test.__mro__[2]]  # pylint: disable=unidiomatic-typecheck

            assert len(candidates) == 1

            out_values.append((candidates[0], type_under_test))

        return out_values

    @pytest.fixture(name="value_type", params=["as_value", "as_generator"])
    def generator_or_no(self, request, type_under_test: type) -> type | None:
        return type_under_test if request.param == "as_generator" else None

    @pytest.fixture(name="value")
    def lazy_value(self, non_lazy_value, request: pytest.FixtureRequest, template_context):
        value_type = None  # DTFIX-FUTURE: get from request when needed, can't easily add to the static fixture list- the generator cases are not being tested
        if value_type:
            if isinstance(non_lazy_value, c.Mapping):
                generator = ((k, v) for k, v in non_lazy_value.items())
            else:
                generator = (item for item in non_lazy_value)

            value = generator

        else:
            value = _AnsibleLazyTemplateMixin._try_create(non_lazy_value)

        # any tests not marked `pytest.mark.allow_delazify` will supply a lazy with its _templar removed and assert that it's still empty afterward
        allow_delazify = any(request.node.iter_markers('allow_delazify'))

        if isinstance(value, _AnsibleLazyTemplateMixin) and not allow_delazify:
            assert value._templar
            # supply a non-functional, but non-None templar, forcing an error if lazy behavior is triggered during tagging
            value._templar = object()  # type: ignore[assignment]

        yield value  # yield to the test; we'll validate later

        # LazyAccessTuple can't template, so we can't induce this failure
        if isinstance(value, _AnsibleLazyTemplateMixin) and not allow_delazify and not isinstance(value, _AnsibleLazyAccessTuple):
            with pytest.raises(AttributeError):
                # verify using the templar fails by using a method which relies on it (to ensure our templar hack above worked)
                t.cast(AnsibleTaggedObject, value)._native_copy()



from __future__ import annotations

import ansible.plugins.connection.local as ansible_local
from ansible.errors import AnsibleConnectionFailure

from ansible.utils.display import Display
display = Display()


class Connection(ansible_local.Connection):
    def exec_command(self, cmd, in_data=None, sudoable=True):
        display.debug('Intercepted call to exec remote command')
        raise AnsibleConnectionFailure('BADLOCAL Error: this is supposed to fail')



#!/usr/bin/python
from __future__ import annotations


DOCUMENTATION = """
---
module: test_docs_removed_precedence
short_description: Test module
description:
    - Test module
author:
    - Ansible Core Team
deprecated:
  alternative: new_module
  why: Updated module released with more functionality
  removed_at_date: '2022-06-01'
  removed_in: '2.14'
"""

EXAMPLES = """
"""

RETURN = """
"""


from ansible.module_utils.basic import AnsibleModule


def main():
    module = AnsibleModule(
        argument_spec=dict(),
    )

    module.exit_json()


if __name__ == '__main__':
    main()



# (c) 2012, Michael DeHaan <michael.dehaan@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from ansible.plugins.action import ActionBase


class ActionModule(ActionBase):
    TRANSFERS_FILES = False

    def run(self, tmp=None, task_vars=None):
        if task_vars is None:
            task_vars = dict()

        if self._task.environment and any(self._task.environment):
            self._display.warning('raw module does not support the environment keyword')

        result = super(ActionModule, self).run(tmp, task_vars)
        del tmp  # tmp no longer has any effect

        if self._task.check_mode:
            # in --check mode, always skip this module execution
            result['skipped'] = True
            return result

        executable = self._task.args.get('executable', False)
        result.update(self._low_level_execute_command(self._task.args.get('_raw_params'), executable=executable))

        result['changed'] = True

        if 'rc' in result and result['rc'] != 0:
            result['failed'] = True
            result['msg'] = 'non-zero return code'

        return result



"""
Disallow direct linking to S3 buckets.
S3 buckets should be accessed through a CloudFront distribution.
"""

from __future__ import annotations

import re
import sys


def main():
    """Main entry point."""
    for path in sys.argv[1:] or sys.stdin.read().splitlines():
        with open(path, 'rb') as path_fd:
            for line, b_text in enumerate(path_fd.readlines()):
                try:
                    text = b_text.decode()
                except UnicodeDecodeError:
                    continue

                if match := re.search(r'(http.*?s3\..*?amazonaws\.com)', text):
                    print(f'{path}:{line + 1}:{match.start(1) + 1}: use a CloudFront distribution instead of an S3 bucket: {match.group(1)}')


if __name__ == '__main__':
    main()



"""Check shebangs, execute bits and byte order marks."""

from __future__ import annotations

import os
import re
import stat
import sys


def main():
    """Main entry point."""
    standard_shebangs = set([
        b'#!/bin/bash -eu',
        b'#!/bin/bash -eux',
        b'#!/bin/sh',
        b'#!/usr/bin/env bash',
        b'#!/usr/bin/env fish',
        b'#!/usr/bin/env pwsh',
        b'#!/usr/bin/env python',
        b'#!/usr/bin/make -f',
    ])

    integration_shebangs = set([
        b'#!/bin/sh',
        b'#!/usr/bin/env bash',
        b'#!/usr/bin/env python',
    ])

    module_shebangs = {
        '': b'#!/usr/bin/python',
        '.py': b'#!/usr/bin/python',
        '.ps1': b'#!powershell',
    }

    # see https://unicode.org/faq/utf_bom.html#bom1
    byte_order_marks = (
        (b'\x00\x00\xFE\xFF', 'UTF-32 (BE)'),
        (b'\xFF\xFE\x00\x00', 'UTF-32 (LE)'),
        (b'\xFE\xFF', 'UTF-16 (BE)'),
        (b'\xFF\xFE', 'UTF-16 (LE)'),
        (b'\xEF\xBB\xBF', 'UTF-8'),
    )

    for path in sys.argv[1:] or sys.stdin.read().splitlines():
        with open(path, 'rb') as path_fd:
            shebang = path_fd.readline().strip()
            mode = os.stat(path).st_mode
            executable = (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH) & mode

            if not shebang or not shebang.startswith(b'#!'):
                if executable:
                    print('%s:%d:%d: file without shebang should not be executable' % (path, 0, 0))

                for mark, name in byte_order_marks:
                    if shebang.startswith(mark):
                        print('%s:%d:%d: file starts with a %s byte order mark' % (path, 0, 0, name))
                        break

                continue

            is_module = False
            is_integration = False

            dirname = os.path.dirname(path)

            if path.startswith('lib/ansible/modules/'):
                is_module = True
            elif re.search('^test/support/[^/]+/plugins/modules/', path):
                is_module = True
            elif re.search('^test/support/[^/]+/collections/ansible_collections/[^/]+/[^/]+/plugins/modules/', path):
                is_module = True
            elif path == 'test/lib/ansible_test/_util/target/cli/ansible_test_cli_stub.py':
                pass  # ansible-test entry point must be executable and have a shebang
            elif re.search(r'^lib/ansible/cli/[^/]+\.py', path):
                pass  # cli entry points must be executable and have a shebang
            elif path.startswith('examples/'):
                continue  # examples trigger some false positives due to location
            elif path.startswith('lib/') or path.startswith('test/lib/'):
                if executable:
                    print('%s:%d:%d: should not be executable' % (path, 0, 0))

                if shebang:
                    print('%s:%d:%d: should not have a shebang' % (path, 0, 0))

                continue
            elif path.startswith('test/integration/targets/') or path.startswith('tests/integration/targets/'):
                is_integration = True

                if dirname.endswith('/library') or '/plugins/modules' in dirname or dirname in (
                        # non-standard module library directories
                        'test/integration/targets/module_precedence/lib_no_extension',
                        'test/integration/targets/module_precedence/lib_with_extension',
                ):
                    is_module = True
            elif path.startswith('plugins/modules/'):
                is_module = True

            if is_module:
                if executable:
                    print('%s:%d:%d: module should not be executable' % (path, 0, 0))

                ext = os.path.splitext(path)[1]
                expected_shebang = module_shebangs.get(ext)
                expected_ext = ' or '.join(['"%s"' % k for k in module_shebangs])

                if expected_shebang:
                    if shebang == expected_shebang:
                        continue

                    print('%s:%d:%d: expected module shebang "%s" but found: %s' % (path, 1, 1, expected_shebang, shebang))
                else:
                    print('%s:%d:%d: expected module extension %s but found: %s' % (path, 0, 0, expected_ext, ext))
            else:
                if is_integration:
                    allowed = integration_shebangs
                else:
                    allowed = standard_shebangs

                if shebang not in allowed:
                    print('%s:%d:%d: unexpected non-module shebang: %s' % (path, 1, 1, shebang))


if __name__ == '__main__':
    main()



# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

from ansible.module_utils.facts.network import linux

# ip -4 route show table local
IP4_ROUTE_SHOW_LOCAL = """
broadcast 127.0.0.0 dev lo proto kernel scope link src 127.0.0.1
local 127.0.0.0/8 dev lo proto kernel scope host src 127.0.0.1
local 127.0.0.1 dev lo proto kernel scope host src 127.0.0.1
broadcast 127.255.255.255 dev lo proto kernel scope link src 127.0.0.1
local 192.168.1.0/24 dev lo scope host
"""

# ip -6 route show table local
IP6_ROUTE_SHOW_LOCAL = """
local ::1 dev lo proto kernel metric 0 pref medium
local 2a02:123:3:1::e dev enp94s0f0np0 proto kernel metric 0 pref medium
local 2a02:123:15::/48 dev lo metric 1024 pref medium
local 2a02:123:16::/48 dev lo metric 1024 pref medium
local fe80::2eea:7fff:feca:fe68 dev enp94s0f0np0 proto kernel metric 0 pref medium
multicast ff00::/8 dev enp94s0f0np0 proto kernel metric 256 pref medium
"""

# Hash returned by get_locally_reachable_ips()
IP_ROUTE_SHOW_LOCAL_EXPECTED = {
    'ipv4': [
        '127.0.0.0/8',
        '127.0.0.1',
        '192.168.1.0/24'
    ],
    'ipv6': [
        '::1',
        '2a02:123:3:1::e',
        '2a02:123:15::/48',
        '2a02:123:16::/48',
        'fe80::2eea:7fff:feca:fe68'
    ]
}


def mock_get_bin_path(command):
    cmds = {"ip": "fake/ip"}
    return cmds.get(command, None)


def test_linux_local_routes(mocker):
    module = mocker.MagicMock()
    mocker.patch.object(module, "get_bin_path", side_effect=mock_get_bin_path)
    mocker.patch.object(
        module,
        "run_command",
        side_effect=[(0, IP4_ROUTE_SHOW_LOCAL, ""), (0, IP6_ROUTE_SHOW_LOCAL, "")],
    )

    net = linux.LinuxNetwork(module)
    res = net.get_locally_reachable_ips(mock_get_bin_path("ip"))
    assert res == IP_ROUTE_SHOW_LOCAL_EXPECTED



# (c) 2021 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import os
import typing as t

from ansible.module_utils.facts.collector import BaseFactCollector


class LoadAvgFactCollector(BaseFactCollector):
    name = 'loadavg'
    _fact_ids = set()  # type: t.Set[str]

    def collect(self, module=None, collected_facts=None):
        facts = {}
        try:
            # (0.58, 0.82, 0.98)
            loadavg = os.getloadavg()
            facts['loadavg'] = {
                '1m': loadavg[0],
                '5m': loadavg[1],
                '15m': loadavg[2]
            }
        except OSError:
            pass

        return facts



# -*- coding: utf-8 -*-
# Copyright (c) 2025, Felix Fontein <felix@fontein.de>, The Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations


DOCUMENTATION = r"""
name: broken
short_description: Test input precedence
author: Felix Fontein (@felixfontein)
description:
  - Test input precedence.
options:
  _terms:
    description:
      - Ignored.
    type: list
    elements: str
    required: true
  some_option:
    description:
      - The interesting part.
    type: str
    default: default value
    env:
      - name: PLAYGROUND_TEST_1
      - name: PLAYGROUND_TEST_2
    vars:
      - name: playground_test_1
      - name: playground_test_2
    ini:
      - key: playground_test_1
        section: playground
      - key: playground_test_2
        section: playground
"""

EXAMPLES = r"""#"""

RETURN = r"""
_list:
  description:
    - The value of O(some_option).
  type: list
  elements: str
"""

from ansible.plugins.lookup import LookupBase


class LookupModule(LookupBase):
    def run(self, terms, variables=None, **kwargs):
        """Generate list."""
        self.set_options(var_options=variables, direct=kwargs)

        return [self.get_option("some_option"), *self.get_option_and_origin("some_option")]



from __future__ import annotations

import copy
import dataclasses
import typing as t

from ansible._internal._templating._jinja_common import validate_arg_type
from ansible._internal._templating._lazy_containers import _AnsibleLazyTemplateMixin
from ansible._internal._templating._transform import _type_transform_mapping
from ansible.errors import AnsibleError


def unmask(value: object, type_names: str | list[str]) -> object:
    """
    Internal filter to suppress automatic type transformation in Jinja (e.g., WarningSummary, DeprecationSummary, ErrorSummary).
    Lazy collection caching is in play - the first attempt to access a value in a given lazy container must be with unmasking in place, or the transformed value
    will already be cached.
    """
    validate_arg_type("type_names", type_names, (str, list))

    if isinstance(type_names, str):
        check_type_names = [type_names]
    else:
        check_type_names = type_names

    valid_type_names = {key.__name__ for key in _type_transform_mapping}
    invalid_type_names = [type_name for type_name in check_type_names if type_name not in valid_type_names]

    if invalid_type_names:
        raise AnsibleError(f'Unknown type name(s): {", ".join(invalid_type_names)}', obj=type_names)

    result: object

    if isinstance(value, _AnsibleLazyTemplateMixin):
        result = copy.copy(value)
        result._lazy_options = dataclasses.replace(
            result._lazy_options,
            unmask_type_names=result._lazy_options.unmask_type_names | frozenset(check_type_names),
        )
    else:
        result = value

    return result


class FilterModule(object):
    @staticmethod
    def filters() -> dict[str, t.Callable]:
        return dict(unmask=unmask)



from __future__ import annotations

import time
import threading
import os

from concurrent.futures import ThreadPoolExecutor, wait

from ansible._internal import _locking


def test_blocking_lock(tmp_path: str) -> None:
    """Validate synchronization behavior of `named_mutex`."""
    work_count = 0
    running_threads: set[int] = set()

    def exercise_lock() -> None:
        nonlocal work_count

        with _locking.named_mutex(os.path.join(tmp_path, "ansible-core")):
            thread_id = threading.current_thread().native_id

            try:
                assert not running_threads

                running_threads.add(thread_id)
                work_count += 1
                time.sleep(0.1)
            finally:
                running_threads.remove(thread_id)

    with ThreadPoolExecutor(max_workers=10) as executor:
        jobs = [executor.submit(exercise_lock) for _idx in range(10)]

        wait(jobs, 5)

        assert all(job.result() is None for job in jobs)
        assert work_count == 10



# Copyright (c) 2019 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = """
    name: test
    plugin_type: inventory
    short_description: test inventory source
    extends_documentation_fragment:
        - inventory_cache
"""

from ansible.plugins.inventory import BaseInventoryPlugin, Cacheable
from ansible.template import trust_as_template


class InventoryModule(BaseInventoryPlugin, Cacheable):

    NAME = 'test'

    def populate(self, hosts):
        for host in list(hosts.keys()):
            self.inventory.add_host(host, group='all')

            for hostvar, hostval in hosts[host].items():
                self.inventory.set_variable(host, hostvar, hostval)

    def get_hosts(self):
        return dict(
            host1=dict(
                one='two',
                my_template=trust_as_template("{{ one }}"),
                verify='two',
            ),
            host2=dict(
                three='four',
                my_template=trust_as_template("{{ three }}"),
                verify='four',
            ),
        )

    def verify_file(self, path):
        return path.endswith('.inventoryconfig.yml')

    def parse(self, inventory, loader, path, cache=True):
        super(InventoryModule, self).parse(inventory, loader, path)

        self.load_cache_plugin()

        cache_key = self.get_cache_key(path)

        # cache may be True or False at this point to indicate if the inventory is being refreshed
        # get the user's cache option
        cache_setting = self.get_option('cache')

        attempt_to_read_cache = cache_setting and cache
        cache_needs_update = cache_setting and not cache

        results = {}

        # attempt to read the cache if inventory isn't being refreshed and the user has caching enabled
        if attempt_to_read_cache:
            try:
                results = self._cache[cache_key]
            except KeyError:
                # This occurs if the cache_key is not in the cache or if the cache_key expired, so the cache needs to be updated
                cache_needs_update = True

        if cache_needs_update:
            results = self.get_hosts()

            # set the cache
            self._cache[cache_key] = results

        self.populate(results)



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Ansible Project
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)

from __future__ import annotations as _annotations

import typing as _t

from ansible.module_utils._internal import _traceback, _deprecator, _event_utils, _messages, _errors
from ansible.module_utils import _internal


def warn(
    warning: str,
    *,
    help_text: str | None = None,
    obj: object | None = None,
) -> None:
    """Record a warning to be returned with the module result."""
    _skip_stackwalk = True

    if _internal.is_controller:
        _display = _internal.import_controller_module('ansible.utils.display').Display()
        _display.warning(
            msg=warning,
            help_text=help_text,
            obj=obj,
        )

        return

    warning = _messages.WarningSummary(
        event=_messages.Event(
            msg=warning,
            help_text=help_text,
            formatted_traceback=_traceback.maybe_capture_traceback(warning, _traceback.TracebackEvent.WARNING),
        ),
    )

    _global_warnings[warning] = None


def error_as_warning(
    msg: str | None,
    exception: BaseException,
    *,
    help_text: str | None = None,
    obj: object = None,
) -> None:
    """Display an exception as a warning."""
    _skip_stackwalk = True

    if _internal.is_controller:
        _display = _internal.import_controller_module('ansible.utils.display').Display()
        _display.error_as_warning(
            msg=msg,
            exception=exception,
            help_text=help_text,
            obj=obj,
        )

        return

    event = _errors.EventFactory.from_exception(exception, _traceback.is_traceback_enabled(_traceback.TracebackEvent.WARNING))

    warning = _messages.WarningSummary(
        event=_messages.Event(
            msg=msg,
            help_text=help_text,
            formatted_traceback=_traceback.maybe_capture_traceback(msg, _traceback.TracebackEvent.WARNING),
            chain=_messages.EventChain(
                msg_reason=_errors.MSG_REASON_DIRECT_CAUSE,
                traceback_reason=_errors.TRACEBACK_REASON_EXCEPTION_DIRECT_WARNING,
                event=event,
            ),
        ),
    )

    _global_warnings[warning] = None


def deprecate(
    msg: str,
    version: str | None = None,
    date: str | None = None,
    collection_name: str | None = None,
    *,
    deprecator: _messages.PluginInfo | None = None,
    help_text: str | None = None,
    obj: object | None = None,
) -> None:
    """
    Record a deprecation warning.
    The `obj` argument is only useful in a controller context; it is ignored for target-side callers.
    Most callers do not need to provide `collection_name` or `deprecator` -- but provide only one if needed.
    Specify `version` or `date`, but not both.
    If `date` is a string, it must be in the form `YYYY-MM-DD`.
    """
    _skip_stackwalk = True

    deprecator = _deprecator.get_best_deprecator(deprecator=deprecator, collection_name=collection_name)

    if _internal.is_controller:
        _display = _internal.import_controller_module('ansible.utils.display').Display()
        _display.deprecated(
            msg=msg,
            version=version,
            date=date,
            help_text=help_text,
            obj=obj,
            # skip passing collection_name; get_best_deprecator already accounted for it when present
            deprecator=deprecator,
        )

        return

    warning = _messages.DeprecationSummary(
        event=_messages.Event(
            msg=msg,
            help_text=help_text,
            formatted_traceback=_traceback.maybe_capture_traceback(msg, _traceback.TracebackEvent.DEPRECATED),
        ),
        version=version,
        date=date,
        deprecator=deprecator,
    )

    _global_deprecations[warning] = None


def get_warning_messages() -> tuple[str, ...]:
    """Return a tuple of warning messages accumulated over this run."""
    # DTFIX7: add future deprecation comment
    return tuple(_event_utils.format_event_brief_message(item.event) for item in _global_warnings)


def get_deprecation_messages() -> tuple[dict[str, _t.Any], ...]:
    """Return a tuple of deprecation warning messages accumulated over this run."""
    # DTFIX7: add future deprecation comment
    return tuple(_event_utils.deprecation_as_dict(item) for item in _global_deprecations)


def get_warnings() -> list[_messages.WarningSummary]:
    """Return a list of warning messages accumulated over this run."""
    return list(_global_warnings)


def get_deprecations() -> list[_messages.DeprecationSummary]:
    """Return a list of deprecations accumulated over this run."""
    return list(_global_deprecations)


_global_warnings: dict[_messages.WarningSummary, object] = {}
"""Global, ordered, de-duplicated storage of accumulated warnings for the current module run."""

_global_deprecations: dict[_messages.DeprecationSummary, object] = {}
"""Global, ordered, de-duplicated storage of accumulated deprecations for the current module run."""



# -*- coding: utf-8 -*-
# Copyright: (c) 2023, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""Tests for depresolver dataclass objects."""


from __future__ import annotations

import pytest

from ansible.galaxy.dependency_resolution.dataclasses import Requirement


NO_LEADING_WHITESPACES = pytest.mark.xfail(
    reason='Does not yet support leading whitespaces',
    strict=True,
)


@pytest.mark.parametrize(
    ('collection_version_spec', 'expected_is_pinned_outcome'),
    (
        ('1.2.3-dev4', True),
        (' 1.2.3-dev4', True),
        ('=1.2.3', True),
        ('= 1.2.3', True),
        (' = 1.2.3', True),
        (' =1.2.3', True),
        ('==1.2.3', True),
        ('== 1.2.3', True),
        (' == 1.2.3', True),
        (' ==1.2.3', True),
        ('!=1.0.0', False),
        ('!= 1.0.0', False),
        pytest.param(' != 1.0.0', False, marks=NO_LEADING_WHITESPACES),
        pytest.param(' !=1.0.0', False, marks=NO_LEADING_WHITESPACES),
        ('>1.0.0', False),
        ('> 1.0.0', False),
        pytest.param(' > 1.0.0', False, marks=NO_LEADING_WHITESPACES),
        pytest.param(' >1.0.0', False, marks=NO_LEADING_WHITESPACES),
        ('>=1.0.0', False),
        ('>= 1.0.0', False),
        pytest.param(' >= 1.0.0', False, marks=NO_LEADING_WHITESPACES),
        pytest.param(' >=1.0.0', False, marks=NO_LEADING_WHITESPACES),
        ('<1.0.0', False),
        ('< 1.0.0', False),
        pytest.param(' < 1.0.0', False, marks=NO_LEADING_WHITESPACES),
        pytest.param(' <1.0.0', False, marks=NO_LEADING_WHITESPACES),
        ('*', False),
        ('* ', False),
        pytest.param(' * ', False, marks=NO_LEADING_WHITESPACES),
        pytest.param(' *', False, marks=NO_LEADING_WHITESPACES),
        ('=1.2.3,!=1.2.3rc5', True),
    ),
)
def test_requirement_is_pinned_logic(
        collection_version_spec: str,
        expected_is_pinned_outcome: bool,
) -> None:
    """Test how Requirement's is_pinned property detects pinned spec."""
    assert Requirement(
        'namespace.collection', collection_version_spec,
        None, None, None,
    ).is_pinned is expected_is_pinned_outcome



"""Become abstraction for interacting with test hosts."""

from __future__ import annotations

import abc
import shlex

from .util import (
    get_subclasses,
)


class Become(metaclass=abc.ABCMeta):
    """Base class for become implementations."""

    @classmethod
    def name(cls) -> str:
        """The name of this plugin."""
        return cls.__name__.lower()

    @property
    @abc.abstractmethod
    def method(self) -> str:
        """The name of the Ansible become plugin that is equivalent to this."""

    @abc.abstractmethod
    def prepare_command(self, command: list[str]) -> list[str]:
        """Return the given command, if any, with privilege escalation."""


class Doas(Become):
    """Become using 'doas'."""

    @property
    def method(self) -> str:
        """The name of the Ansible become plugin that is equivalent to this."""
        raise NotImplementedError('Ansible has no built-in doas become plugin.')

    def prepare_command(self, command: list[str]) -> list[str]:
        """Return the given command, if any, with privilege escalation."""
        become = ['doas', '-n']

        if command:
            become.extend(['sh', '-c', shlex.join(command)])
        else:
            become.extend(['-s'])

        return become


class DoasSudo(Doas):
    """Become using 'doas' in ansible-test and then after bootstrapping use 'sudo' for other ansible commands."""

    @classmethod
    def name(cls) -> str:
        """The name of this plugin."""
        return 'doas_sudo'

    @property
    def method(self) -> str:
        """The name of the Ansible become plugin that is equivalent to this."""
        return 'sudo'


class Su(Become):
    """Become using 'su'."""

    @property
    def method(self) -> str:
        """The name of the Ansible become plugin that is equivalent to this."""
        return 'su'

    def prepare_command(self, command: list[str]) -> list[str]:
        """Return the given command, if any, with privilege escalation."""
        become = ['su', '-l', 'root']

        if command:
            become.extend(['-c', shlex.join(command)])

        return become


class SuSudo(Su):
    """Become using 'su' in ansible-test and then after bootstrapping use 'sudo' for other ansible commands."""

    @classmethod
    def name(cls) -> str:
        """The name of this plugin."""
        return 'su_sudo'

    @property
    def method(self) -> str:
        """The name of the Ansible become plugin that is equivalent to this."""
        return 'sudo'


class Sudo(Become):
    """Become using 'sudo'."""

    @property
    def method(self) -> str:
        """The name of the Ansible become plugin that is equivalent to this."""
        return 'sudo'

    def prepare_command(self, command: list[str]) -> list[str]:
        """Return the given command, if any, with privilege escalation."""
        become = ['sudo', '-in']

        if command:
            become.extend(['sh', '-c', shlex.join(command)])

        return become


SUPPORTED_BECOME_METHODS = {cls.name(): cls for cls in get_subclasses(Become)}



"""Timeout management for tests."""

from __future__ import annotations

import dataclasses
import datetime
import functools
import os
import signal
import time
import typing as t

from .io import (
    read_json_file,
)

from .config import (
    CommonConfig,
    TestConfig,
)

from .util import (
    display,
    TimeoutExpiredError,
)

from .thread import (
    WrappedThread,
)

from .constants import (
    TIMEOUT_PATH,
)

from .test import (
    TestTimeout,
)


@dataclasses.dataclass(frozen=True)
class TimeoutDetail:
    """Details required to enforce a timeout on test execution."""

    _DEADLINE_FORMAT = '%Y-%m-%dT%H:%M:%SZ'  # format used to maintain backwards compatibility with previous versions of ansible-test

    deadline: datetime.datetime
    duration: int | float  # minutes

    @property
    def remaining(self) -> datetime.timedelta:
        """The amount of time remaining before the timeout occurs. If the timeout has passed, this will be a negative duration."""
        return self.deadline - datetime.datetime.now(tz=datetime.timezone.utc).replace(microsecond=0)

    def to_dict(self) -> dict[str, t.Any]:
        """Return timeout details as a dictionary suitable for JSON serialization."""
        return dict(
            deadline=self.deadline.strftime(self._DEADLINE_FORMAT),
            duration=self.duration,
        )

    @staticmethod
    def from_dict(value: dict[str, t.Any]) -> TimeoutDetail:
        """Return a TimeoutDetail instance using the value previously returned by to_dict."""
        return TimeoutDetail(
            deadline=datetime.datetime.strptime(value['deadline'], TimeoutDetail._DEADLINE_FORMAT).replace(tzinfo=datetime.timezone.utc),
            duration=value['duration'],
        )

    @staticmethod
    def create(duration: int | float) -> TimeoutDetail | None:
        """Return a new TimeoutDetail instance for the specified duration (in minutes), or None if the duration is zero."""
        if not duration:
            return None

        if duration == int(duration):
            duration = int(duration)

        return TimeoutDetail(
            deadline=datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0) + datetime.timedelta(seconds=int(duration * 60)),
            duration=duration,
        )


def get_timeout() -> TimeoutDetail | None:
    """Return details about the currently set timeout, if any, otherwise return None."""
    try:
        return TimeoutDetail.from_dict(read_json_file(TIMEOUT_PATH))
    except FileNotFoundError:
        return None


def configure_timeout(args: CommonConfig) -> None:
    """Configure the timeout."""
    if isinstance(args, TestConfig):
        configure_test_timeout(args)  # only tests are subject to the timeout


def configure_test_timeout(args: TestConfig) -> None:
    """Configure the test timeout."""
    timeout = get_timeout()

    if not timeout:
        return

    timeout_remaining = timeout.remaining

    test_timeout = TestTimeout(timeout.duration)

    if timeout_remaining <= datetime.timedelta():
        test_timeout.write(args)

        raise TimeoutExpiredError(f'The {timeout.duration} minute test timeout expired {timeout_remaining * -1} ago at {timeout.deadline}.')

    display.info(f'The {timeout.duration} minute test timeout expires in {timeout_remaining} at {timeout.deadline}.', verbosity=1)

    def timeout_handler(_dummy1: t.Any, _dummy2: t.Any) -> None:
        """Runs when SIGUSR1 is received."""
        test_timeout.write(args)

        raise TimeoutExpiredError(f'Tests aborted after exceeding the {timeout.duration} minute time limit.')

    def timeout_waiter(timeout_seconds: float) -> None:
        """Background thread which will kill the current process if the timeout elapses."""
        time.sleep(timeout_seconds)
        os.kill(os.getpid(), signal.SIGUSR1)

    signal.signal(signal.SIGUSR1, timeout_handler)

    instance = WrappedThread(functools.partial(timeout_waiter, timeout_remaining.total_seconds()), 'Timeout Watchdog')
    instance.daemon = True
    instance.start()



from __future__ import annotations

from ansible.plugins.lookup import LookupBase
from ansible.utils.display import Display


class LookupModule(LookupBase):
    def run(self, terms, variables=None, **kwargs):
        Display().deprecated("Hello World!", version='2.9999')

        return []



# (c) 2014 Michael DeHaan, <michael@ansible.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from ansible.errors import AnsibleError
from ansible.module_utils.six import string_types
from ansible.playbook.role.definition import RoleDefinition
from ansible.utils.display import Display
from ansible.utils.galaxy import scm_archive_resource

__all__ = ['RoleRequirement']

VALID_SPEC_KEYS = [
    'name',
    'role',
    'scm',
    'src',
    'version',
]

display = Display()


class RoleRequirement(RoleDefinition):

    """
    Helper class for Galaxy, which is used to parse both dependencies
    specified in meta/main.yml and requirements.yml files.
    """

    def __init__(self):
        pass

    @staticmethod
    def repo_url_to_role_name(repo_url):
        # gets the role name out of a repo like
        # http://git.example.com/repos/repo.git" => "repo"

        if '://' not in repo_url and '@' not in repo_url:
            return repo_url
        trailing_path = repo_url.split('/')[-1]
        if trailing_path.endswith('.git'):
            trailing_path = trailing_path[:-4]
        if trailing_path.endswith('.tar.gz'):
            trailing_path = trailing_path[:-7]
        if ',' in trailing_path:
            trailing_path = trailing_path.split(',')[0]
        return trailing_path

    @staticmethod
    def role_yaml_parse(role):

        if isinstance(role, string_types):
            name = None
            scm = None
            src = None
            version = None
            if ',' in role:
                if role.count(',') == 1:
                    (src, version) = role.strip().split(',', 1)
                elif role.count(',') == 2:
                    (src, version, name) = role.strip().split(',', 2)
                else:
                    raise AnsibleError("Invalid role line (%s). Proper format is 'role_name[,version[,name]]'" % role)
            else:
                src = role

            if name is None:
                name = RoleRequirement.repo_url_to_role_name(src)
            if '+' in src:
                (scm, src) = src.split('+', 1)

            return dict(name=name, src=src, scm=scm, version=version)

        if 'role' in role:
            name = role['role']
            if ',' in name:
                raise AnsibleError("Invalid old style role requirement: %s" % name)
            else:
                del role['role']
                role['name'] = name
        else:
            role = role.copy()

            if 'src' in role:
                # New style: { src: 'galaxy.role,version,name', other_vars: "here" }
                if 'github.com' in role["src"] and 'http' in role["src"] and '+' not in role["src"] and not role["src"].endswith('.tar.gz'):
                    role["src"] = "git+" + role["src"]

                if '+' in role["src"]:
                    role["scm"], dummy, role["src"] = role["src"].partition('+')

                if 'name' not in role:
                    role["name"] = RoleRequirement.repo_url_to_role_name(role["src"])

            if 'version' not in role:
                role['version'] = ''

            if 'scm' not in role:
                role['scm'] = None

        for key in list(role.keys()):
            if key not in VALID_SPEC_KEYS:
                role.pop(key)

        return role

    @staticmethod
    def scm_archive_role(src, scm='git', name=None, version='HEAD', keep_scm_meta=False):

        return scm_archive_resource(src, scm=scm, name=name, version=version, keep_scm_meta=keep_scm_meta)



# -*- coding: utf-8 -*-
# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

# For Backward compatibility
from ansible.module_utils.common.sentinel import Sentinel  # pylint: disable=unused-import



from __future__ import annotations

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.datatag import deprecate_value

METADATA = """
schema_version: 1
serialization_profile: modern
"""


def main():
    mod = AnsibleModule(argument_spec={
        'sensitive_module_arg': dict(default='THIS VALUE WAS NO_LOG IN A MODULE AND SHOULD NOT BE SEEN', type='str', no_log=True),
    })

    something_old_value = 'an old thing'
    # Deprecated needs args; tag the value and store it
    something_old_value = deprecate_value(something_old_value, "`something_old` is deprecated, don't use it!", version='9.9999')

    result = {
        'something_old': something_old_value,
        # send the module param back to core
        'sensitive_module_arg': mod.params['sensitive_module_arg'],
        # rendering templates from modules is a no-no, core does not trust anything by default
        'untrusted_template': '{{ ["me", "see", "not", "should"] | sort(reverse=true) | join(" ") }}',
        'changed': False
    }

    mod.exit_json(**result)


if __name__ == '__main__':
    main()



from __future__ import annotations

import pathlib

from ansible._internal._templating._utils import Omit
from ansible.parsing.dataloader import DataLoader
from ansible.template import Templar, trust_as_template


def test_no_finalize_marker_passthru(tmp_path: pathlib.Path) -> None:
    """Return an Undefined marker from a template lookup to ensure that the internal templating operation does not finalize its result."""
    template_path = tmp_path / 'template.txt'
    template_path.write_text("{{ bogusvar }}")

    templar = Templar(loader=DataLoader(), variables=dict(template_path=str(template_path)))

    assert templar.template(trust_as_template('{{ lookup("template", template_path) | default("pass") }}')) == "pass"


def test_no_finalize_omit_passthru(tmp_path: pathlib.Path) -> None:
    """Return an Omit scalar from a template lookup to ensure that the internal templating operation does not finalize its result."""
    template_path = tmp_path / 'template.txt'
    template_path.write_text("{{ omitted }}")

    data = dict(omitted=trust_as_template("{{ omit }}"), template_path=str(template_path))

    # The result from the lookup should be an Omit value, since the result of the template lookup's internal templating call should not be finalized.
    # If it were, finalize would trip the Omit and raise an error about a top-level template result resolving to an Omit scalar.
    res = Templar(loader=DataLoader(), variables=data).template(trust_as_template("{{ lookup('template', template_path) | type_debug }}"))

    assert res == type(Omit).__name__



from __future__ import annotations

import abc
import collections.abc as c
import typing as t

from yaml.nodes import ScalarNode, Node

from ansible._internal._templating import _jinja_common
from ansible.module_utils import _internal
from ansible.module_utils._internal._datatag import AnsibleTaggedObject, Tripwire, AnsibleTagHelper
from ansible.parsing.vault import VaultHelper
from ansible.module_utils.common.yaml import HAS_LIBYAML

if HAS_LIBYAML:
    from yaml.cyaml import CSafeDumper as SafeDumper
else:
    from yaml import SafeDumper  # type: ignore[assignment]


class _BaseDumper(SafeDumper, metaclass=abc.ABCMeta):
    """Base class for Ansible YAML dumpers."""

    @classmethod
    @abc.abstractmethod
    def _register_representers(cls) -> None:
        """Method used to register representers to derived types during class initialization."""

    def __init_subclass__(cls, **kwargs) -> None:
        """Initialization for derived types."""
        cls._register_representers()


class AnsibleDumper(_BaseDumper):
    """A simple stub class that allows us to add representers for our custom types."""

    @classmethod
    def _register_representers(cls) -> None:
        cls.add_multi_representer(AnsibleTaggedObject, cls.represent_ansible_tagged_object)
        cls.add_multi_representer(Tripwire, cls.represent_tripwire)
        cls.add_multi_representer(c.Mapping, cls.represent_dict)
        cls.add_multi_representer(c.Collection, cls.represent_list)
        cls.add_multi_representer(_jinja_common.VaultExceptionMarker, cls.represent_vault_exception_marker)

    def get_node_from_ciphertext(self, data: object) -> ScalarNode | None:
        if ciphertext := VaultHelper.get_ciphertext(data, with_tags=False):
            return self.represent_scalar('!vault', ciphertext, style='|')

        return None

    def represent_vault_exception_marker(self, data: _jinja_common.VaultExceptionMarker) -> ScalarNode:
        if node := self.get_node_from_ciphertext(data):
            return node

        data.trip()

    def represent_ansible_tagged_object(self, data: AnsibleTaggedObject) -> Node:
        if _internal.is_intermediate_mapping(data):
            return self.represent_dict(data)

        if _internal.is_intermediate_iterable(data):
            return self.represent_list(data)

        if node := self.get_node_from_ciphertext(data):
            return node

        return self.represent_data(AnsibleTagHelper.as_native_type(data))  # automatically decrypts encrypted strings

    def represent_tripwire(self, data: Tripwire) -> t.NoReturn:
        data.trip()



from __future__ import annotations


def yolo(value):
    return True


class TestModule(object):
    """ Ansible core jinja2 tests """

    def tests(self):
        return {
            # failure testing
            'yolo': yolo,
        }



# WARNING: Changing line numbers of code in this file will break collection tests that use tracing to check paths and line numbers.
#          Also, do not import annotations from __future__ as this will break detection of __future__ inheritance.


def question() -> float:
    return 3 / 2



# -*- coding: utf-8 -*-
# (c) 2015, Toshio Kuratomi <tkuratomi@ansible.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import warnings

try:
    # deprecated: description='warning suppression only required for Python 3.12 and earlier' python_version='3.12'
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message="'crypt' is deprecated and slated for removal in Python 3.13", category=DeprecationWarning)

        import passlib
        from passlib.handlers import pbkdf2
except ImportError:  # pragma: nocover
    passlib = None
    pbkdf2 = None

import pytest

from units.mock.loader import DictDataLoader

import unittest
from unittest.mock import mock_open, patch
from ansible.errors import AnsibleError
import builtins
from ansible.module_utils.common.text.converters import to_bytes
from ansible.plugins.loader import PluginLoader, lookup_loader
from ansible.plugins.lookup import password


DEFAULT_LENGTH = 20
DEFAULT_CHARS = sorted([u'ascii_letters', u'digits', u".,:-_"])
DEFAULT_CANDIDATE_CHARS = u'.,:-_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'

# Currently there isn't a new-style
old_style_params_data = (
    # Simple case
    dict(
        term=u'/path/to/file',
        filename=u'/path/to/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None, chars=DEFAULT_CHARS, seed=None),
        candidate_chars=DEFAULT_CANDIDATE_CHARS,
    ),

    # Special characters in path
    dict(
        term=u'/path/with/embedded spaces and/file',
        filename=u'/path/with/embedded spaces and/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None, chars=DEFAULT_CHARS, seed=None),
        candidate_chars=DEFAULT_CANDIDATE_CHARS,
    ),
    dict(
        term=u'/path/with/equals/cn=com.ansible',
        filename=u'/path/with/equals/cn=com.ansible',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None, chars=DEFAULT_CHARS, seed=None),
        candidate_chars=DEFAULT_CANDIDATE_CHARS,
    ),
    dict(
        term=u'/path/with/unicode/くらとみ/file',
        filename=u'/path/with/unicode/くらとみ/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None, chars=DEFAULT_CHARS, seed=None),
        candidate_chars=DEFAULT_CANDIDATE_CHARS,
    ),

    # Mix several special chars
    dict(
        term=u'/path/with/utf 8 and spaces/くらとみ/file',
        filename=u'/path/with/utf 8 and spaces/くらとみ/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None, chars=DEFAULT_CHARS, seed=None),
        candidate_chars=DEFAULT_CANDIDATE_CHARS,
    ),
    dict(
        term=u'/path/with/encoding=unicode/くらとみ/file',
        filename=u'/path/with/encoding=unicode/くらとみ/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None, chars=DEFAULT_CHARS, seed=None),
        candidate_chars=DEFAULT_CANDIDATE_CHARS,
    ),
    dict(
        term=u'/path/with/encoding=unicode/くらとみ/and spaces file',
        filename=u'/path/with/encoding=unicode/くらとみ/and spaces file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None, chars=DEFAULT_CHARS, seed=None),
        candidate_chars=DEFAULT_CANDIDATE_CHARS,
    ),

    # Simple parameters
    dict(
        term=u'/path/to/file length=42',
        filename=u'/path/to/file',
        params=dict(length=42, encrypt=None, ident=None, chars=DEFAULT_CHARS, seed=None),
        candidate_chars=DEFAULT_CANDIDATE_CHARS,
    ),
    dict(
        term=u'/path/to/file encrypt=pbkdf2_sha256',
        filename=u'/path/to/file',
        params=dict(length=DEFAULT_LENGTH, encrypt='pbkdf2_sha256', ident=None, chars=DEFAULT_CHARS, seed=None),
        candidate_chars=DEFAULT_CANDIDATE_CHARS,
    ),
    dict(
        term=u'/path/to/file chars=abcdefghijklmnop',
        filename=u'/path/to/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None, chars=[u'abcdefghijklmnop'], seed=None),
        candidate_chars=u'abcdefghijklmnop',
    ),
    dict(
        term=u'/path/to/file chars=digits,abc,def',
        filename=u'/path/to/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None,
                    chars=sorted([u'digits', u'abc', u'def']), seed=None),
        candidate_chars=u'abcdef0123456789',
    ),
    dict(
        term=u'/path/to/file seed=1',
        filename=u'/path/to/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None, chars=DEFAULT_CHARS, seed='1'),
        candidate_chars=DEFAULT_CANDIDATE_CHARS,
    ),

    # Including comma in chars
    dict(
        term=u'/path/to/file chars=abcdefghijklmnop,,digits',
        filename=u'/path/to/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None,
                    chars=sorted([u'abcdefghijklmnop', u',', u'digits']), seed=None),
        candidate_chars=u',abcdefghijklmnop0123456789',
    ),
    dict(
        term=u'/path/to/file chars=,,',
        filename=u'/path/to/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None,
                    chars=[u','], seed=None),
        candidate_chars=u',',
    ),

    # Including = in chars
    dict(
        term=u'/path/to/file chars=digits,=,,',
        filename=u'/path/to/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None,
                    chars=sorted([u'digits', u'=', u',']), seed=None),
        candidate_chars=u',=0123456789',
    ),
    dict(
        term=u'/path/to/file chars=digits,abc=def',
        filename=u'/path/to/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None,
                    chars=sorted([u'digits', u'abc=def']), seed=None),
        candidate_chars=u'abc=def0123456789',
    ),

    # Including unicode in chars
    dict(
        term=u'/path/to/file chars=digits,くらとみ,,',
        filename=u'/path/to/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None,
                    chars=sorted([u'digits', u'くらとみ', u',']), seed=None),
        candidate_chars=u',0123456789くらとみ',
    ),
    # Including only unicode in chars
    dict(
        term=u'/path/to/file chars=くらとみ',
        filename=u'/path/to/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None,
                    chars=sorted([u'くらとみ']), seed=None),
        candidate_chars=u'くらとみ',
    ),

    # Include ':' in path
    dict(
        term=u'/path/to/file_with:colon chars=ascii_letters,digits',
        filename=u'/path/to/file_with:colon',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None,
                    chars=sorted([u'ascii_letters', u'digits']), seed=None),
        candidate_chars=u'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789',
    ),

    # Including special chars in both path and chars
    # Special characters in path
    dict(
        term=u'/path/with/embedded spaces and/file chars=abc=def',
        filename=u'/path/with/embedded spaces and/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None, chars=[u'abc=def'], seed=None),
        candidate_chars=u'abc=def',
    ),
    dict(
        term=u'/path/with/equals/cn=com.ansible chars=abc=def',
        filename=u'/path/with/equals/cn=com.ansible',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None, chars=[u'abc=def'], seed=None),
        candidate_chars=u'abc=def',
    ),
    dict(
        term=u'/path/with/unicode/くらとみ/file chars=くらとみ',
        filename=u'/path/with/unicode/くらとみ/file',
        params=dict(length=DEFAULT_LENGTH, encrypt=None, ident=None, chars=[u'くらとみ'], seed=None),
        candidate_chars=u'くらとみ',
    ),
)


class TestParseParameters(unittest.TestCase):

    def setUp(self):
        self.fake_loader = DictDataLoader({'/path/to/somewhere': 'sdfsdf'})
        self.password_lookup = lookup_loader.get('password')
        self.password_lookup._loader = self.fake_loader

    def test(self):
        for testcase in old_style_params_data:
            filename, params = self.password_lookup._parse_parameters(testcase['term'])
            params['chars'].sort()
            self.assertEqual(filename, testcase['filename'])
            self.assertEqual(params, testcase['params'])

    def test_unrecognized_value(self):
        testcase = dict(term=u'/path/to/file chars=くらとみi  sdfsdf',
                        filename=u'/path/to/file',
                        params=dict(length=DEFAULT_LENGTH, encrypt=None, chars=[u'くらとみ']),
                        candidate_chars=u'くらとみ')
        self.assertRaises(AnsibleError, self.password_lookup._parse_parameters, testcase['term'])

    def test_invalid_params(self):
        testcase = dict(term=u'/path/to/file chars=くらとみi  somethign_invalid=123',
                        filename=u'/path/to/file',
                        params=dict(length=DEFAULT_LENGTH, encrypt=None, chars=[u'くらとみ']),
                        candidate_chars=u'くらとみ')
        self.assertRaises(AnsibleError, self.password_lookup._parse_parameters, testcase['term'])


class TestReadPasswordFile(unittest.TestCase):
    def setUp(self):
        self.os_path_exists = password.os.path.exists

    def tearDown(self):
        password.os.path.exists = self.os_path_exists

    def test_no_password_file(self):
        password.os.path.exists = lambda x: False
        self.assertEqual(password._read_password_file(b'/nonexistent'), None)

    def test_with_password_file(self):
        password.os.path.exists = lambda x: True
        with patch.object(builtins, 'open', mock_open(read_data=b'Testing\n')) as m:
            self.assertEqual(password._read_password_file(b'/etc/motd'), u'Testing')


class TestGenCandidateChars(unittest.TestCase):
    def _assert_gen_candidate_chars(self, testcase):
        expected_candidate_chars = testcase['candidate_chars']
        params = testcase['params']
        chars_spec = params['chars']
        res = password._gen_candidate_chars(chars_spec)
        self.assertEqual(res, expected_candidate_chars)

    def test_gen_candidate_chars(self):
        for testcase in old_style_params_data:
            self._assert_gen_candidate_chars(testcase)


class TestRandomPassword(unittest.TestCase):
    def _assert_valid_chars(self, res, chars):
        for res_char in res:
            self.assertIn(res_char, chars)

    def test_default(self):
        res = password.random_password()
        self.assertEqual(len(res), DEFAULT_LENGTH)
        self.assertTrue(isinstance(res, str))
        self._assert_valid_chars(res, DEFAULT_CANDIDATE_CHARS)

    def test_zero_length(self):
        res = password.random_password(length=0)
        self.assertEqual(len(res), 0)
        self.assertTrue(isinstance(res, str))
        self._assert_valid_chars(res, u',')

    def test_just_a_common(self):
        res = password.random_password(length=1, chars=u',')
        self.assertEqual(len(res), 1)
        self.assertEqual(res, u',')

    def test_free_will(self):
        # A Rush and Spinal Tap reference twofer
        res = password.random_password(length=11, chars=u'a')
        self.assertEqual(len(res), 11)
        self.assertEqual(res, 'aaaaaaaaaaa')
        self._assert_valid_chars(res, u'a')

    def test_unicode(self):
        res = password.random_password(length=11, chars=u'くらとみ')
        self._assert_valid_chars(res, u'くらとみ')
        self.assertEqual(len(res), 11)

    def test_seed(self):
        pw1 = password.random_password(seed=1)
        pw2 = password.random_password(seed=1)
        pw3 = password.random_password(seed=2)
        self.assertEqual(pw1, pw2)
        self.assertNotEqual(pw1, pw3)

    def test_gen_password(self):
        for testcase in old_style_params_data:
            params = testcase['params']
            candidate_chars = testcase['candidate_chars']
            params_chars_spec = password._gen_candidate_chars(params['chars'])
            password_string = password.random_password(length=params['length'],
                                                       chars=params_chars_spec)
            self.assertEqual(len(password_string),
                             params['length'],
                             msg='generated password=%s has length (%s) instead of expected length (%s)' %
                             (password_string, len(password_string), params['length']))

            for char in password_string:
                self.assertIn(char, candidate_chars,
                              msg='%s not found in %s from chars spect %s' %
                              (char, candidate_chars, params['chars']))


class TestParseContent(unittest.TestCase):

    def test_empty_password_file(self):
        plaintext_password, salt, ident = password._parse_content(u'')
        self.assertEqual(plaintext_password, u'')
        self.assertEqual(salt, None)
        self.assertEqual(ident, None)

    def test(self):
        expected_content = u'12345678'
        file_content = expected_content
        plaintext_password, salt, ident = password._parse_content(file_content)
        self.assertEqual(plaintext_password, expected_content)
        self.assertEqual(salt, None)
        self.assertEqual(ident, None)

    def test_with_salt(self):
        expected_content = u'12345678 salt=87654321'
        file_content = expected_content
        plaintext_password, salt, ident = password._parse_content(file_content)
        self.assertEqual(plaintext_password, u'12345678')
        self.assertEqual(salt, u'87654321')
        self.assertEqual(ident, None)

    def test_with_salt_and_ident(self):
        expected_content = u'12345678 salt=87654321 ident=2a'
        file_content = expected_content
        plaintext_password, salt, ident = password._parse_content(file_content)
        self.assertEqual(plaintext_password, u'12345678')
        self.assertEqual(salt, u'87654321')
        self.assertEqual(ident, u'2a')


class TestFormatContent(unittest.TestCase):
    def test_no_encrypt(self):
        self.assertEqual(
            password._format_content(password=u'hunter42',
                                     salt=u'87654321',
                                     encrypt=False),
            u'hunter42 salt=87654321')

    def test_no_encrypt_no_salt(self):
        self.assertEqual(
            password._format_content(password=u'hunter42',
                                     salt=None,
                                     encrypt=None),
            u'hunter42')

    def test_encrypt(self):
        self.assertEqual(
            password._format_content(password=u'hunter42',
                                     salt=u'87654321',
                                     encrypt='pbkdf2_sha256'),
            u'hunter42 salt=87654321')

    def test_encrypt_no_salt(self):
        self.assertRaises(AssertionError, password._format_content, u'hunter42', None, 'pbkdf2_sha256')


class TestWritePasswordFile(unittest.TestCase):
    def setUp(self):
        self.makedirs_safe = password.makedirs_safe
        self.os_chmod = password.os.chmod
        password.makedirs_safe = self.noop
        password.os.chmod = self.noop

    def noop(self, *args, **kwargs):
        pass

    def tearDown(self):
        password.makedirs_safe = self.makedirs_safe
        password.os.chmod = self.os_chmod

    def test_content_written(self):

        with patch.object(builtins, 'open', mock_open()) as m:
            password._write_password_file(b'/this/is/a/test/caf\xc3\xa9', u'Testing Café')

            m.assert_called_once_with(b'/this/is/a/test/caf\xc3\xa9', 'wb')
            m().write.assert_called_once_with(u'Testing Café\n'.encode('utf-8'))


class BaseTestLookupModule(unittest.TestCase):
    def setUp(self):
        self.fake_loader = DictDataLoader({'/path/to/somewhere': 'sdfsdf'})
        self.password_lookup = lookup_loader.get('password')
        self.password_lookup._loader = self.fake_loader
        self.os_path_exists = password.os.path.exists
        self.os_open = password.os.open
        password.os.open = self.noop
        self.os_close = password.os.close
        password.os.close = self.noop
        self.makedirs_safe = password.makedirs_safe
        password.makedirs_safe = self.noop

    def noop(self, *args, **kwargs):
        pass

    def tearDown(self):
        password.os.path.exists = self.os_path_exists
        password.os.open = self.os_open
        password.os.close = self.os_close
        password.makedirs_safe = self.makedirs_safe


class TestLookupModuleWithoutPasslib(BaseTestLookupModule):
    @patch.object(PluginLoader, '_get_paths')
    @patch('ansible.plugins.lookup.password._write_password_file')
    def test_no_encrypt(self, mock_get_paths, mock_write_file):
        mock_get_paths.return_value = ['/path/one', '/path/two', '/path/three']

        results = self.password_lookup.run([u'/path/to/somewhere'], None)

        # FIXME: assert something useful
        for result in results:
            assert len(result) == DEFAULT_LENGTH
            assert isinstance(result, str)

    @patch.object(PluginLoader, '_get_paths')
    @patch('ansible.plugins.lookup.password._write_password_file')
    def test_password_already_created_no_encrypt(self, mock_get_paths, mock_write_file):
        mock_get_paths.return_value = ['/path/one', '/path/two', '/path/three']
        password.os.path.exists = lambda x: x == to_bytes('/path/to/somewhere')

        with patch.object(builtins, 'open', mock_open(read_data=b'hunter42 salt=87654321\n')) as m:
            results = self.password_lookup.run([u'/path/to/somewhere chars=anything'], None)

        for result in results:
            self.assertEqual(result, u'hunter42')

    @patch.object(PluginLoader, '_get_paths')
    @patch('ansible.plugins.lookup.password._write_password_file')
    def test_only_a(self, mock_get_paths, mock_write_file):
        mock_get_paths.return_value = ['/path/one', '/path/two', '/path/three']

        results = self.password_lookup.run([u'/path/to/somewhere chars=a'], None)
        for result in results:
            self.assertEqual(result, u'a' * DEFAULT_LENGTH)

    @patch('time.sleep')
    def test_lock_been_held(self, mock_sleep):
        # pretend the lock file is here
        def _already_exists(*args, **kwargs):
            raise FileExistsError("The lock is busy, wait and try again.")

        with (
            pytest.raises(AnsibleError, match='^Password lookup cannot get the lock in 7 seconds.*'),
            patch.object(password.os, 'open', _already_exists),
            patch.object(password.os.path, 'exists', lambda *args, **kwargs: True),
        ):
            # should timeout here
            self.password_lookup.run([u'/path/to/somewhere chars=anything'], None)

    def test_lock_not_been_held(self):
        # pretend now there is password file but no lock
        password.os.path.exists = lambda x: x == to_bytes('/path/to/somewhere')
        with patch.object(builtins, 'open', mock_open(read_data=b'hunter42 salt=87654321\n')) as m:
            # should not timeout here
            results = self.password_lookup.run([u'/path/to/somewhere chars=anything'], None)

        for result in results:
            self.assertEqual(result, u'hunter42')


@pytest.mark.skipif(passlib is None, reason='passlib must be installed to run these tests')
class TestLookupModuleWithPasslib(BaseTestLookupModule):
    def setUp(self):
        super(TestLookupModuleWithPasslib, self).setUp()

        # Different releases of passlib default to a different number of rounds
        self.sha256 = passlib.registry.get_crypt_handler('pbkdf2_sha256')
        sha256_for_tests = pbkdf2.create_pbkdf2_hash("sha256", 32, 20000)
        passlib.registry.register_crypt_handler(sha256_for_tests, force=True)

    def tearDown(self):
        super(TestLookupModuleWithPasslib, self).tearDown()

        passlib.registry.register_crypt_handler(self.sha256, force=True)

    @patch.object(PluginLoader, '_get_paths')
    @patch('ansible.plugins.lookup.password._write_password_file')
    def test_encrypt(self, mock_get_paths, mock_write_file):
        mock_get_paths.return_value = ['/path/one', '/path/two', '/path/three']

        results = self.password_lookup.run([u'/path/to/somewhere encrypt=pbkdf2_sha256'], None)

        # pbkdf2 format plus hash
        expected_password_length = 76

        for result in results:
            self.assertEqual(len(result), expected_password_length)
            # result should have 5 parts split by '$'
            str_parts = result.split('$', 5)

            # verify the result is parseable by the passlib
            crypt_parts = passlib.hash.pbkdf2_sha256.parsehash(result)

            # verify it used the right algo type
            self.assertEqual(str_parts[1], 'pbkdf2-sha256')

            self.assertEqual(len(str_parts), 5)

            # verify the string and parsehash agree on the number of rounds
            self.assertEqual(int(str_parts[2]), crypt_parts['rounds'])
            self.assertIsInstance(result, str)

    @patch('ansible.plugins.lookup.password._write_password_file')
    def test_password_already_created_encrypt(self, mock_write_file):
        password.os.path.exists = lambda x: x == to_bytes('/path/to/somewhere')

        with patch.object(builtins, 'open', mock_open(read_data=b'hunter42 salt=87654321\n')) as m:
            results = self.password_lookup.run([u'/path/to/somewhere chars=anything encrypt=pbkdf2_sha256'], None)
        for result in results:
            self.assertEqual(result, u'$pbkdf2-sha256$20000$ODc2NTQzMjE$Uikde0cv0BKaRaAXMrUQB.zvG4GmnjClwjghwIRf2gU')

        # Assert the password file is not rewritten
        mock_write_file.assert_not_called()


@pytest.mark.skipif(passlib is None, reason='passlib must be installed to run these tests')
class TestLookupModuleWithPasslibWrappedAlgo(BaseTestLookupModule):
    def setUp(self):
        super(TestLookupModuleWithPasslibWrappedAlgo, self).setUp()
        self.os_path_exists = password.os.path.exists

    def tearDown(self):
        super(TestLookupModuleWithPasslibWrappedAlgo, self).tearDown()
        password.os.path.exists = self.os_path_exists

    @patch('ansible.plugins.lookup.password._write_password_file')
    def test_encrypt_wrapped_crypt_algo(self, mock_write_file):

        password.os.path.exists = self.password_lookup._loader.path_exists
        with patch.object(builtins, 'open', mock_open(read_data=self.password_lookup._loader.get_text_file_contents('/path/to/somewhere'))):
            results = self.password_lookup.run([u'/path/to/somewhere encrypt=ldap_sha256_crypt'], None)

            wrapper = getattr(passlib.hash, 'ldap_sha256_crypt')

            self.assertEqual(len(results), 1)
            result = results[0]
            self.assertIsInstance(result, str)

            expected_password_length = 76
            self.assertEqual(len(result), expected_password_length)

            # result should have 5 parts split by '$'
            str_parts = result.split('$')
            self.assertEqual(len(str_parts), 5)

            # verify the string and passlib agree on the number of rounds
            self.assertEqual(str_parts[2], "rounds=%s" % wrapper.default_rounds)

            # verify it used the right algo type
            self.assertEqual(str_parts[0], '{CRYPT}')

            # verify it used the right algo type
            self.assertTrue(wrapper.verify(self.password_lookup._loader.get_text_file_contents('/path/to/somewhere'), result))

            # verify a password with a non default rounds value
            # generated with: echo test | mkpasswd -s --rounds 660000 -m sha-256 --salt testansiblepass.
            hashpw = '{CRYPT}$5$rounds=660000$testansiblepass.$KlRSdA3iFXoPI.dEwh7AixiXW3EtCkLrlQvlYA2sluD'
            self.assertTrue(wrapper.verify('test', hashpw))



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

DOCUMENTATION = """
    name: linear
    short_description: Executes tasks in a linear fashion
    description:
        - Task execution is in lockstep per host batch as defined by C(serial) (default all).
          Up to the fork limit of hosts will execute each task at the same time and then
          the next series of hosts until the batch is done, before going on to the next task.
    version_added: "2.0"
    notes:
     - This was the default Ansible behaviour before 'strategy plugins' were introduced in 2.0.
    author: Ansible Core Team
"""

from ansible import constants as C
from ansible.errors import AnsibleError, AnsibleAssertionError, AnsibleParserError, AnsibleValueOmittedError
from ansible.playbook.handler import Handler
from ansible.playbook.included_file import IncludedFile
from ansible.plugins.loader import action_loader
from ansible.plugins.strategy import StrategyBase
from ansible._internal._templating._engine import TemplateEngine
from ansible.utils.display import Display
from ansible.inventory.host import Host
from ansible.playbook.task import Task
from ansible.executor.play_iterator import PlayIterator
from ansible.playbook.play_context import PlayContext
from ansible.executor import task_result as _task_result

display = Display()


class StrategyModule(StrategyBase):

    def _get_next_task_lockstep(self, hosts: list[Host], iterator: PlayIterator) -> list[tuple[Host, Task]]:
        """
        Returns a list of (host, task) tuples, where the task may
        be a noop task to keep the iterator in lock step across
        all hosts.
        """

        state_task_per_host = {}
        for host in hosts:
            state, task = iterator.get_next_task_for_host(host, peek=True)
            if task is not None:
                state_task_per_host[host] = state, task

        if not state_task_per_host:
            return []

        task_uuids = {t._uuid for s, t in state_task_per_host.values()}
        _loop_cnt = 0
        while _loop_cnt <= 1:
            try:
                cur_task = iterator.all_tasks[iterator.cur_task]
            except IndexError:
                # pick up any tasks left after clear_host_errors
                iterator.cur_task = 0
                _loop_cnt += 1
            else:
                iterator.cur_task += 1
                if cur_task._uuid in task_uuids:
                    break
        else:
            # prevent infinite loop
            raise AnsibleAssertionError(
                'BUG: There seems to be a mismatch between tasks in PlayIterator and HostStates.'
            )

        host_tasks = []
        for host, (state, task) in state_task_per_host.items():
            if cur_task._uuid == task._uuid:
                iterator.set_state_for_host(host.name, state)
                host_tasks.append((host, task))

        if cur_task._get_meta() == 'flush_handlers':
            iterator.all_tasks[iterator.cur_task:iterator.cur_task] = [h for b in iterator._play.handlers for h in b.block]

        return host_tasks

    def run(self, iterator, play_context: PlayContext):  # type: ignore[override]
        """
        The linear strategy is simple - get the next task and queue
        it for all hosts, then wait for the queue to drain before
        moving on to the next task
        """

        # iterate over each task, while there is one left to run
        result = int(self._tqm.RUN_OK)
        work_to_do = True

        self._set_hosts_cache(iterator._play)

        while work_to_do and not self._tqm._terminated:

            try:
                display.debug("getting the remaining hosts for this loop")
                hosts_left = self.get_hosts_left(iterator)
                display.debug("done getting the remaining hosts for this loop")

                # queue up this task for each host in the inventory
                callback_sent = False
                work_to_do = False

                host_tasks = self._get_next_task_lockstep(hosts_left, iterator)

                # skip control
                skip_rest = False
                choose_step = True

                # flag set if task is set to any_errors_fatal
                any_errors_fatal = False

                results: list[_task_result._RawTaskResult] = []
                for (host, task) in host_tasks:
                    if self._tqm._terminated:
                        break

                    run_once = False
                    work_to_do = True

                    host_name = host.get_name()

                    display.debug("getting variables")
                    task_vars = self._variable_manager.get_vars(play=iterator._play, host=host, task=task,
                                                                _hosts=self._hosts_cache, _hosts_all=self._hosts_cache_all)
                    self.add_tqm_variables(task_vars, play=iterator._play)
                    templar = TemplateEngine(loader=self._loader, variables=task_vars)
                    display.debug("done getting variables")

                    # test to see if the task across all hosts points to an action plugin which
                    # sets BYPASS_HOST_LOOP to true, or if it has run_once enabled. If so, we
                    # will only send this task to the first host in the list.

                    try:
                        task_action = templar.template(task.action)
                    except AnsibleValueOmittedError:
                        raise AnsibleParserError("Omit is not valid for the `action` keyword.", obj=task.action) from None

                    try:
                        action = action_loader.get(task_action, class_only=True, collection_list=task.collections)
                    except KeyError:
                        # we don't care here, because the action may simply not have a
                        # corresponding action plugin
                        action = None

                    if task_action in C._ACTION_META:
                        # for the linear strategy, we run meta tasks just once and for
                        # all hosts currently being iterated over rather than one host
                        results.extend(self._execute_meta(task, play_context, iterator, host))
                        if task._get_meta() not in ('noop', 'reset_connection', 'end_host', 'role_complete', 'flush_handlers', 'end_role'):
                            run_once = True
                        if (task.any_errors_fatal or run_once) and not task.ignore_errors:
                            any_errors_fatal = True
                    else:
                        # handle step if needed, skip meta actions as they are used internally
                        if self._step and choose_step:
                            if self._take_step(task):
                                choose_step = False
                            else:
                                skip_rest = True
                                break

                        run_once = action and getattr(action, 'BYPASS_HOST_LOOP', False) or templar.template(task.run_once)

                        if (task.any_errors_fatal or run_once) and not task.ignore_errors:
                            any_errors_fatal = True

                        if not callback_sent:
                            task.post_validate_attribute("name", templar=templar)

                            if isinstance(task, Handler):
                                self._tqm.send_callback('v2_playbook_on_handler_task_start', task)
                            else:
                                self._tqm.send_callback('v2_playbook_on_task_start', task, is_conditional=False)
                            callback_sent = True

                        self._blocked_hosts[host_name] = True
                        self._queue_task(host, task, task_vars, play_context)
                        del task_vars

                    if isinstance(task, Handler):
                        if run_once:
                            task.clear_hosts()
                        else:
                            task.remove_host(host)

                    # if we're bypassing the host loop, break out now
                    if run_once:
                        break

                    results.extend(self._process_pending_results(iterator, max_passes=max(1, int(len(self._tqm._workers) * 0.1))))

                # go to next host/task group
                if skip_rest:
                    continue

                display.debug("done queuing things up, now waiting for results queue to drain")
                if self._pending_results > 0:
                    results.extend(self._wait_on_pending_results(iterator))

                self.update_active_connections(results)

                included_files = IncludedFile.process_include_results(
                    results,
                    iterator=iterator,
                    loader=self._loader,
                    variable_manager=self._variable_manager
                )

                if len(included_files) > 0:
                    display.debug("we have included files to process")

                    display.debug("generating all_blocks data")
                    all_blocks = dict((host, []) for host in hosts_left)
                    display.debug("done generating all_blocks data")
                    included_tasks = []
                    failed_includes_hosts = set()
                    for included_file in included_files:
                        display.debug("processing included file: %s" % included_file._filename)
                        is_handler = False
                        try:
                            if included_file._is_role:
                                new_ir = self._copy_included_file(included_file)

                                new_blocks, handler_blocks = new_ir.get_block_list(
                                    play=iterator._play,
                                    variable_manager=self._variable_manager,
                                    loader=self._loader,
                                )
                            else:
                                is_handler = isinstance(included_file._task, Handler)
                                new_blocks = self._load_included_file(
                                    included_file,
                                    iterator=iterator,
                                    is_handler=is_handler,
                                    handle_stats_and_callbacks=False,
                                )

                            # let PlayIterator know about any new handlers included via include_role or
                            # import_role within include_role/include_taks
                            iterator.handlers = [h for b in iterator._play.handlers for h in b.block]

                            display.debug("iterating over new_blocks loaded from include file")
                            for new_block in new_blocks:
                                if is_handler:
                                    for task in new_block.block:
                                        task.notified_hosts = included_file._hosts[:]
                                    final_block = new_block
                                else:
                                    task_vars = self._variable_manager.get_vars(
                                        play=iterator._play,
                                        task=new_block.get_first_parent_include(),
                                        _hosts=self._hosts_cache,
                                        _hosts_all=self._hosts_cache_all,
                                    )
                                    display.debug("filtering new block on tags")
                                    final_block = new_block.filter_tagged_tasks(task_vars)
                                    display.debug("done filtering new block on tags")

                                included_tasks.extend(final_block.get_tasks())

                                for host in hosts_left:
                                    if host in included_file._hosts:
                                        all_blocks[host].append(final_block)

                            display.debug("done iterating over new_blocks loaded from include file")
                        except AnsibleParserError:
                            raise
                        except AnsibleError as ex:
                            # FIXME: send the error to the callback; don't directly write to display here
                            display.error(ex)
                            for r in included_file._results:
                                r._return_data['failed'] = True
                                r._return_data['reason'] = str(ex)
                                self._tqm._stats.increment('failures', r.host.name)
                                self._tqm.send_callback('v2_runner_on_failed', r)
                                failed_includes_hosts.add(r.host)
                        else:
                            # since we skip incrementing the stats when the task result is
                            # first processed, we do so now for each host in the list
                            for host in included_file._hosts:
                                self._tqm._stats.increment('ok', host.name)
                            self._tqm.send_callback('v2_playbook_on_include', included_file)

                    for host in failed_includes_hosts:
                        self._tqm._failed_hosts[host.name] = True
                        iterator.mark_host_failed(host)

                    # finally go through all of the hosts and append the
                    # accumulated blocks to their list of tasks
                    display.debug("extending task lists for all hosts with included blocks")

                    for host in hosts_left:
                        iterator.add_tasks(host, all_blocks[host])

                    iterator.all_tasks[iterator.cur_task:iterator.cur_task] = included_tasks

                    display.debug("done extending task lists")
                    display.debug("done processing included files")

                display.debug("results queue empty")

                display.debug("checking for any_errors_fatal")
                failed_hosts = []
                unreachable_hosts = []
                for res in results:
                    if res.is_failed():
                        failed_hosts.append(res.host.name)
                    elif res.is_unreachable():
                        unreachable_hosts.append(res.host.name)

                if any_errors_fatal and (failed_hosts or unreachable_hosts):
                    for host in hosts_left:
                        if host.name not in failed_hosts:
                            self._tqm._failed_hosts[host.name] = True
                            iterator.mark_host_failed(host)
                display.debug("done checking for any_errors_fatal")

                display.debug("checking for max_fail_percentage")
                if iterator._play.max_fail_percentage is not None and len(results) > 0:
                    percentage = iterator._play.max_fail_percentage / 100.0

                    if (len(self._tqm._failed_hosts) / iterator.batch_size) > percentage:
                        for host in hosts_left:
                            # don't double-mark hosts, or the iterator will potentially
                            # fail them out of the rescue/always states
                            if host.name not in failed_hosts:
                                self._tqm._failed_hosts[host.name] = True
                                iterator.mark_host_failed(host)
                        self._tqm.send_callback('v2_playbook_on_no_hosts_remaining')
                        result |= self._tqm.RUN_FAILED_BREAK_PLAY
                    display.debug('(%s failed / %s total )> %s max fail' % (len(self._tqm._failed_hosts), iterator.batch_size, percentage))
                display.debug("done checking for max_fail_percentage")

                display.debug("checking to see if all hosts have failed and the running result is not ok")
                if result != self._tqm.RUN_OK and len(self._tqm._failed_hosts) >= len(hosts_left):
                    display.debug("^ not ok, so returning result now")
                    self._tqm.send_callback('v2_playbook_on_no_hosts_remaining')
                    return result
                display.debug("done checking to see if all hosts have failed")

            finally:
                # removed unnecessary exception handler, don't want to mis-attribute the entire code block by changing indentation
                pass

        # run the base class run() method, which executes the cleanup function
        # and runs any outstanding handlers which have been triggered

        return super(StrategyModule, self).run(iterator, play_context, result)



from __future__ import annotations

from ansible.plugins.action.normal import ActionModule as NormalAction


class ActionModule(NormalAction):
    def run(self, *args, **kwargs):
        result = super(ActionModule, self).run(*args, **kwargs)
        result['hacked'] = 'I got run under a subclassed normal, yay'
        return result



from __future__ import annotations

import json
import os
import os.path
import pathlib


def main() -> None:
    collection_root = pathlib.Path(os.getcwd())
    print(f"Running windows-integration coverage test in '{collection_root}'")

    result_path = collection_root / "tests" / "output" / "coverage" / "coverage-powershell"
    adjacent_modules_path = collection_root / "tests" / "integration" / "targets" / "win_collection" / "library"
    adjacent_utils_path = collection_root / "tests" / "integration" / "targets" / "win_collection" / "module_utils"
    collection_modules_path = collection_root / "plugins" / "modules"
    collection_utils_path = collection_root / "plugins" / "module_utils"

    expected_hits = {
        str(adjacent_modules_path / 'test_win_collection_async.ps1'): {'5': 1, '6': 1},
        str(adjacent_modules_path / 'test_win_collection_become.ps1'): {'5': 1, '6': 1},
        str(adjacent_modules_path / 'test_win_collection_normal.ps1'): {'6': 1, '7': 1, '8': 1},
        str(adjacent_utils_path / 'Ansible.ModuleUtils.AdjacentPwshCoverage.psm1'): {'6': 1, '9': 1},
        str(collection_modules_path / 'win_collection.ps1'): {'6': 1, '7': 1, '8': 1},
        str(collection_utils_path / 'CollectionPwshCoverage.psm1'): {'6': 1, '9': 1},
    }
    found_hits = set()

    with open(result_path, mode="rb") as fd:
        data = json.load(fd)

    for path, result in data.items():
        print(f"Testing result for path '{path}' -> {result!r}")
        assert path in expected_hits, f"Found unexpected coverage result path '{path}'"

        expected = expected_hits[path]
        assert result == expected, f"Coverage result for {path} was {result!r} but was expecting {expected!r}"
        found_hits.add(path)

    missing_hits = set(expected_hits.keys()).difference(found_hits)
    assert not missing_hits, f"Expected coverage results for {', '.join(missing_hits)} but they were not present"


if __name__ == '__main__':
    main()



# (c) 2012, Michael DeHaan <michael.dehaan@gmail.com>
# (c) 2015, 2017 Toshio Kuratomi <tkuratomi@ansible.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = """
    name: local
    short_description: execute on controller
    description:
        - This connection plugin allows ansible to execute tasks on the Ansible 'controller' instead of on a remote host.
    author: ansible (@core)
    version_added: historical
    options:
        become_success_timeout:
            version_added: '2.19'
            type: int
            default: 10
            description:
                - Number of seconds to wait for become to succeed when enabled.
                - The default will be used if the configured value is less than 1.
            vars:
                - name: ansible_local_become_success_timeout
        become_strip_preamble:
            version_added: '2.19'
            type: bool
            default: true
            description:
                - Strip internal become output preceding command execution. Disable for additional diagnostics.
            vars:
                - name: ansible_local_become_strip_preamble
    extends_documentation_fragment:
        - connection_pipelining
    notes:
        - The remote user is ignored, the user with which the ansible CLI was executed is used instead.
"""

import functools
import getpass
import os
import pty
import selectors
import shutil
import subprocess
import time
import typing as t

import ansible.constants as C
from ansible.errors import AnsibleError, AnsibleFileNotFound, AnsibleConnectionFailure
from ansible.module_utils.six import text_type, binary_type
from ansible.module_utils.common.text.converters import to_bytes, to_native, to_text
from ansible.plugins.connection import ConnectionBase
from ansible.utils.display import Display
from ansible.utils.path import unfrackpath

display = Display()


class Connection(ConnectionBase):
    """ Local based connections """

    transport = 'local'
    has_pipelining = True

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:

        super(Connection, self).__init__(*args, **kwargs)
        self.cwd = None
        try:
            self.default_user = getpass.getuser()
        except KeyError:
            display.vv("Current user (uid=%s) does not seem to exist on this system, leaving user empty." % os.getuid())
            self.default_user = ""

    def _connect(self) -> Connection:
        """ connect to the local host; nothing to do here """

        # Because we haven't made any remote connection we're running as
        # the local user, rather than as whatever is configured in remote_user.
        self._play_context.remote_user = self.default_user

        if not self._connected:
            display.vvv(u"ESTABLISH LOCAL CONNECTION FOR USER: {0}".format(self._play_context.remote_user), host=self._play_context.remote_addr)
            self._connected = True
        return self

    def exec_command(self, cmd: str, in_data: bytes | None = None, sudoable: bool = True) -> tuple[int, bytes, bytes]:
        """ run a command on the local host """

        super(Connection, self).exec_command(cmd, in_data=in_data, sudoable=sudoable)

        display.debug("in local.exec_command()")

        executable = C.DEFAULT_EXECUTABLE.split()[0] if C.DEFAULT_EXECUTABLE else None

        if not os.path.exists(to_bytes(executable, errors='surrogate_or_strict')):
            raise AnsibleError("failed to find the executable specified %s."
                               " Please verify if the executable exists and re-try." % executable)

        display.vvv(u"EXEC {0}".format(to_text(cmd)), host=self._play_context.remote_addr)
        display.debug("opening command with Popen()")

        if isinstance(cmd, (text_type, binary_type)):
            cmd = to_text(cmd)
        else:
            cmd = map(to_text, cmd)

        pty_primary = None
        stdin = subprocess.PIPE
        if sudoable and self.become and self.become.expect_prompt() and not self.get_option('pipelining'):
            # Create a pty if sudoable for privilege escalation that needs it.
            # Falls back to using a standard pipe if this fails, which may
            # cause the command to fail in certain situations where we are escalating
            # privileges or the command otherwise needs a pty.
            try:
                pty_primary, stdin = pty.openpty()
            except OSError as ex:
                display.debug(f"Unable to open pty: {ex}")

        p = subprocess.Popen(
            cmd,
            shell=isinstance(cmd, (text_type, binary_type)),
            executable=executable,
            cwd=self.cwd,
            stdin=stdin,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # if we created a pty, we can close the other half of the pty now, otherwise primary is stdin
        if pty_primary is not None:
            os.close(stdin)

        display.debug("done running command with Popen()")

        become_stdout_bytes, become_stderr_bytes = self._ensure_become_success(p, pty_primary, sudoable)

        display.debug("getting output with communicate()")
        stdout, stderr = p.communicate(in_data)
        display.debug("done communicating")

        # preserve output from privilege escalation stage as `bytes`; it may contain actual output (eg `raw`) or error messages
        stdout = become_stdout_bytes + stdout
        stderr = become_stderr_bytes + stderr

        # finally, close the other half of the pty, if it was created
        if pty_primary:
            os.close(pty_primary)

        display.debug("done with local.exec_command()")
        return p.returncode, stdout, stderr

    def _ensure_become_success(self, p: subprocess.Popen, pty_primary: int, sudoable: bool) -> tuple[bytes, bytes]:
        """
        Ensure that become succeeds, returning a tuple containing stdout captured after success and all stderr.
        Returns immediately if `self.become` or `sudoable` are False, or `build_become_command` was not called, without performing any additional checks.
        """
        if not self.become or not sudoable or not self.become._id:  # _id is set by build_become_command, if it was not called, assume no become
            return b'', b''

        start_seconds = time.monotonic()
        become_stdout = bytearray()
        become_stderr = bytearray()
        last_stdout_prompt_offset = 0
        last_stderr_prompt_offset = 0

        # map the buffers to their associated stream for the selector reads
        become_capture = {
            p.stdout: become_stdout,
            p.stderr: become_stderr,
        }

        expect_password_prompt = self.become.expect_prompt()
        sent_password = False

        def become_error_msg(reason: str) -> str:
            error_message = f'{reason} waiting for become success'

            if expect_password_prompt and not sent_password:
                error_message += ' or become password prompt'

            error_message += '.'

            if become_stdout:
                error_message += f'\n>>> Standard Output\n{to_text(bytes(become_stdout))}'

            if become_stderr:
                error_message += f'\n>>> Standard Error\n{to_text(bytes(become_stderr))}'

            return error_message

        os.set_blocking(p.stdout.fileno(), False)
        os.set_blocking(p.stderr.fileno(), False)

        with selectors.DefaultSelector() as selector:
            selector.register(p.stdout, selectors.EVENT_READ, 'stdout')
            selector.register(p.stderr, selectors.EVENT_READ, 'stderr')

            while not self.become.check_success(become_stdout):
                if not selector.get_map():  # we only reach end of stream after all descriptors are EOF
                    raise AnsibleError(become_error_msg('Premature end of stream'))

                if expect_password_prompt and (
                    self.become.check_password_prompt(bytes(become_stdout[last_stdout_prompt_offset:])) or
                    self.become.check_password_prompt(bytes(become_stderr[last_stderr_prompt_offset:]))
                ):
                    if sent_password:
                        raise AnsibleError(become_error_msg('Duplicate become password prompt encountered'))

                    last_stdout_prompt_offset = len(become_stdout)
                    last_stderr_prompt_offset = len(become_stderr)

                    password_to_send = to_bytes(self.become.get_option('become_pass') or '') + b'\n'

                    if pty_primary is None:
                        p.stdin.write(password_to_send)
                        p.stdin.flush()
                    else:
                        os.write(pty_primary, password_to_send)

                    sent_password = True

                remaining_timeout_seconds = self._become_success_timeout - (time.monotonic() - start_seconds)
                events = selector.select(remaining_timeout_seconds) if remaining_timeout_seconds > 0 else []

                if not events:
                    # ignoring remaining output after timeout to prevent hanging
                    raise AnsibleConnectionFailure(become_error_msg('Timed out'))

                # read all content (non-blocking) from streams that signaled available input and append to the associated buffer
                for key, event in events:
                    obj = t.cast(t.BinaryIO, key.fileobj)

                    if chunk := obj.read():
                        become_capture[obj] += chunk
                    else:
                        selector.unregister(obj)  # EOF on this obj, stop polling it

        os.set_blocking(p.stdout.fileno(), True)
        os.set_blocking(p.stderr.fileno(), True)

        become_stdout_bytes = bytes(become_stdout)
        become_stderr_bytes = bytes(become_stderr)

        if self.get_option('become_strip_preamble'):
            become_stdout_bytes = self.become.strip_become_success(self.become.strip_become_prompt(become_stdout_bytes))
            become_stderr_bytes = self.become.strip_become_prompt(become_stderr_bytes)

        return become_stdout_bytes, become_stderr_bytes

    @functools.cached_property
    def _become_success_timeout(self) -> int:
        """Timeout value for become success in seconds."""
        if (timeout := self.get_option('become_success_timeout')) < 1:
            timeout = C.config.get_config_default('become_success_timeout', plugin_type='connection', plugin_name='local')

        return timeout

    def put_file(self, in_path: str, out_path: str) -> None:
        """ transfer a file from local to local """

        super(Connection, self).put_file(in_path, out_path)

        in_path = unfrackpath(in_path, basedir=self.cwd)
        out_path = unfrackpath(out_path, basedir=self.cwd)

        display.vvv(u"PUT {0} TO {1}".format(in_path, out_path), host=self._play_context.remote_addr)
        if not os.path.exists(to_bytes(in_path, errors='surrogate_or_strict')):
            raise AnsibleFileNotFound("file or module does not exist: {0}".format(to_native(in_path)))
        try:
            shutil.copyfile(to_bytes(in_path, errors='surrogate_or_strict'), to_bytes(out_path, errors='surrogate_or_strict'))
        except shutil.Error:
            raise AnsibleError("failed to copy: {0} and {1} are the same".format(to_native(in_path), to_native(out_path)))
        except OSError as ex:
            raise AnsibleError(f"Failed to transfer file to {out_path!r}.") from ex

    def fetch_file(self, in_path: str, out_path: str) -> None:
        """ fetch a file from local to local -- for compatibility """

        super(Connection, self).fetch_file(in_path, out_path)

        display.vvv(u"FETCH {0} TO {1}".format(in_path, out_path), host=self._play_context.remote_addr)
        self.put_file(in_path, out_path)

    def reset(self) -> None:
        pass

    def close(self) -> None:
        """ terminate the connection; nothing to do here """
        self._connected = False



# (c) 2020 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import functools

from ansible.plugins.callback import CallbackBase


class CallbackModule(CallbackBase):
    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = 'stdout'
    CALLBACK_NAME = 'callback_debug'

    def __init__(self, *args, **kwargs):
        super(CallbackModule, self).__init__(*args, **kwargs)
        self._display.display('__init__')

        for name in (cb for cb in dir(self) if cb.startswith('v2_')):
            setattr(self, name, functools.partial(self.handle_v2, name))

    def handle_v2(self, name, *args, **kwargs):
        self._display.display(name)



#
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The base class for all resource modules
"""
from __future__ import annotations

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.network import (
    get_resource_connection,
)


class ConfigBase(object):
    """ The base class for all resource modules
    """

    ACTION_STATES = ["merged", "replaced", "overridden", "deleted"]

    def __init__(self, module):
        self._module = module
        self.state = module.params["state"]
        self._connection = None

        if self.state not in ["rendered", "parsed"]:
            self._connection = get_resource_connection(module)



#!/usr/bin/python

# Copyright: (c) 2020, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

DOCUMENTATION = r"""
---
module: httptester_kinit
short_description: Get Kerberos ticket
description: Get Kerberos ticket using kinit non-interactively.
options:
  username:
    description: The username to get the ticket for.
    required: true
    type: str
  password:
    description: The password for I(username).
    required; true
    type: str
author:
- Ansible Project
"""

EXAMPLES = r"""
#
"""

RETURN = r"""
#
"""

import contextlib
import os
import subprocess

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_bytes, to_text

try:
    import configparser
except ImportError:
    import ConfigParser as configparser


@contextlib.contextmanager
def env_path(name, value, default_value):
    """ Adds a value to a PATH-like env var and preserve the existing value if present. """
    orig_value = os.environ.get(name, None)
    os.environ[name] = '%s:%s' % (value, orig_value or default_value)
    try:
        yield

    finally:
        if orig_value:
            os.environ[name] = orig_value

        else:
            del os.environ[name]


@contextlib.contextmanager
def krb5_conf(module, config):
    """ Runs with a custom krb5.conf file that extends the existing config if present. """
    if config:
        ini_config = configparser.ConfigParser()
        for section, entries in config.items():
            ini_config.add_section(section)
            for key, value in entries.items():
                ini_config.set(section, key, value)

        config_path = os.path.join(module.tmpdir, 'krb5.conf')
        with open(config_path, mode='wt') as config_fd:
            ini_config.write(config_fd)

        with env_path('KRB5_CONFIG', config_path, '/etc/krb5.conf'):
            yield

    else:
        yield


def main():
    module_args = dict(
        username=dict(type='str', required=True),
        password=dict(type='str', required=True, no_log=True),
    )
    module = AnsibleModule(
        argument_spec=module_args,
        required_together=[('username', 'password')],
    )

    # Heimdal has a few quirks that we want to paper over in this module
    #     1. KRB5_TRACE does not work in any released version (<=7.7), we need to use a custom krb5.config to enable it
    #     2. When reading the password it reads from the pty not stdin by default causing an issue with subprocess. We
    #        can control that behaviour with '--password-file=STDIN'
    # Also need to set the custom path to krb5-config and kinit as FreeBSD relies on the newer Heimdal version in the
    # port package.
    sysname = os.uname()[0]
    prefix = '/usr/local/bin/' if sysname == 'FreeBSD' else ''
    is_heimdal = sysname in ['Darwin', 'FreeBSD']

    # Debugging purposes, get the Kerberos version. On platforms like OpenSUSE this may not be on the PATH.
    try:
        process = subprocess.Popen(['%skrb5-config' % prefix, '--version'], stdout=subprocess.PIPE)
        stdout, stderr = process.communicate()
        version = to_text(stdout)
    except FileNotFoundError:
        version = 'Unknown (no krb5-config)'

    kinit_args = ['%skinit' % prefix]
    config = {}
    if is_heimdal:
        kinit_args.append('--password-file=STDIN')
        config['logging'] = {'krb5': 'FILE:/dev/stdout'}
    kinit_args.append(to_text(module.params['username'], errors='surrogate_or_strict'))

    with krb5_conf(module, config):
        # Weirdly setting KRB5_CONFIG in the modules environment block does not work unless we pass it in explicitly.
        # Take a copy of the existing environment to make sure the process has the same env vars as ours. Also set
        # KRB5_TRACE to output and debug logs helping to identify problems when calling kinit with MIT.
        kinit_env = os.environ.copy()
        kinit_env['KRB5_TRACE'] = '/dev/stdout'

        process = subprocess.Popen(kinit_args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                   env=kinit_env)
        stdout, stderr = process.communicate(to_bytes(module.params['password'], errors='surrogate_or_strict') + b'\n')
        rc = process.returncode

    module.exit_json(changed=True, stdout=to_text(stdout), stderr=to_text(stderr), rc=rc, version=version)


if __name__ == '__main__':
    main()



from __future__ import annotations

import collections.abc as c

import typing as t

if t.TYPE_CHECKING:
    from ansible.module_utils.compat.typing import TypeGuard


INTERMEDIATE_MAPPING_TYPES = (c.Mapping,)
"""
Mapping types which are supported for recursion and runtime usage, such as in serialization and templating.
These will be converted to a simple Python `dict` before serialization or storage as a variable.
"""

INTERMEDIATE_ITERABLE_TYPES = (tuple, set, frozenset, c.Sequence)
"""
Iterable types which are supported for recursion and runtime usage, such as in serialization and templating.
These will be converted to a simple Python `list` before serialization or storage as a variable.
CAUTION: Scalar types which are sequences should be excluded when using this.
"""

ITERABLE_SCALARS_NOT_TO_ITERATE = (str, bytes)
"""Scalars which are also iterable, and should thus be excluded from iterable checks."""


def is_intermediate_mapping(value: object) -> TypeGuard[c.Mapping]:
    """Returns `True` if `value` is a type supported for projection to a Python `dict`, otherwise returns `False`."""
    return isinstance(value, INTERMEDIATE_MAPPING_TYPES)


def is_intermediate_iterable(value: object) -> TypeGuard[c.Iterable]:
    """Returns `True` if `value` is a type supported for projection to a Python `list`, otherwise returns `False`."""
    return isinstance(value, INTERMEDIATE_ITERABLE_TYPES) and not isinstance(value, ITERABLE_SCALARS_NOT_TO_ITERATE)


is_controller: bool = False
"""Set to True automatically when this module is imported into an Ansible controller context."""


def get_controller_serialize_map() -> dict[type, t.Callable]:
    """
    Called to augment serialization maps.
    This implementation is replaced with the one from ansible._internal in controller contexts.
    """
    return {}


def import_controller_module(_module_name: str, /) -> t.Any:
    """
    Called to conditionally import the named module in a controller context, otherwise returns `None`.
    This implementation is replaced with the one from ansible._internal in controller contexts.
    """
    return None



from __future__ import annotations

from ansible_collections.testns.testcoll.plugins.module_utils import secondary
import ansible_collections.testns.testcoll.plugins.module_utils.secondary


def thingtocall():
    if secondary != ansible_collections.testns.testcoll.plugins.module_utils.secondary:
        raise Exception()

    return "thingtocall in base called " + ansible_collections.testns.testcoll.plugins.module_utils.secondary.thingtocall()



from __future__ import annotations

import pathlib

import pytest

from ansible.errors import AnsibleError, AnsibleVariableTypeError
from ansible._internal._errors._error_utils import SourceContext
from ansible._internal._datatag._tags import Origin

from ..test_utils.controller.display import emits_warnings


@pytest.mark.parametrize("filename, line, col, expected_filename", (
    ('nonexistent_file.txt', 3, 4, "nonexistent.txt"),
    ('short_file.txt', None, None, "short_file_no_line.txt"),
    ('short_file.txt', 324762384, 3, "short_file_overflowed_line.txt"),
    ('short_file.txt', 3, 324762384, "short_file_overflowed_col.txt"),
    ('short_file.txt', -1, 3, "short_file_underflowed_line.txt"),
    ('short_file.txt', 3, -1, "short_file_underflowed_col.txt"),
    ('short_file.txt', 5, None, "short_file_no_column.txt"),
    ('short_file.txt', 2, None, "short_file_no_column_overflowed.txt"),
    ('empty_file.txt', 2, 1, "empty_file_unavailable.txt"),
    ('short_file.txt', 1, 1, "short_file_no_context_left_marker.txt"),
    ('short_file.txt', 5, 1, "short_file_left_marker.txt"),
    ('short_file_missing_trailing_newline.txt', 7, 1, "short_file_missing_trailing_newline_left_marker.txt"),
    ('short_file.txt', 7, 118, "short_file_last_column_right_marker.txt"),
    ('short_file.txt', 4, 118, "short_file_truncated_target.txt"),
    ('short_file.txt', 4, 115, "short_file_truncated_target_last_displayed_char.txt"),
    ('short_file.txt', 4, 174, "short_file_long_line_truncated_past_target.txt"),
    ('long_file.txt', 14, 117, "long_file_last_column_right_marker.txt"),
    ('file_with_tabs.txt', 3, 22, "file_with_tabs_replaced_left_marker.txt"),
))
def test_source_context(filename: str, line: int | None, col: int | None, expected_filename: str) -> None:
    """
    Validate various SourceContext formatting scenarios.
    Test case outputs are generated; set `overwrite_expected` to regenerate.
    """
    # set this to true to regenerate and overwrite the expected test output files
    overwrite_expected = False

    fixture_path = pathlib.Path(__file__).parent / 'fixtures'

    origin = Origin(path=str(fixture_path / 'inputs' / filename), line_num=line, col_num=col)

    # DTFIX-FUTURE: assert target_line contents as well?
    source_context = SourceContext.from_origin(origin)
    result = '\n'.join(source_context.annotated_source_lines) + '\n'

    expected_path = fixture_path / 'outputs' / expected_filename

    if overwrite_expected:  # pragma: nocover
        expected_path.write_text(result)

    assert result == expected_path.read_text()
    assert not overwrite_expected


def test_suppress_extended_error_deprecation() -> None:
    with emits_warnings(deprecation_pattern='suppress_extended_error.*argument'):
        AnsibleError(suppress_extended_error=True)  # pylint: disable=pointless-exception-statement


@pytest.mark.parametrize("obj, expected", (
    (1, f'Type {int.__name__!r} is unsupported for variable storage.'),
    (int, f'Type {type.__name__!r} is unsupported for variable storage.'),
))
def test_ansible_variable_type_error(obj: object, expected: str) -> None:
    assert expected in str(AnsibleVariableTypeError.from_value(obj=obj))



from __future__ import annotations

import pytest

from ansible._internal._errors import _error_factory

from ansible.errors import AnsibleError
from ansible._internal import _display_utils
from ansible._internal._datatag._tags import Origin
from ansible._internal._errors._error_utils import format_exception_message
from ansible.module_utils._internal import _messages
from units.mock.error_helper import raise_exceptions


_shared_cause = Exception('shared cause')
_origin_x = Origin(path='/x')
_origin_y = Origin(path='/y')


@pytest.mark.parametrize("exceptions, expected_message_chain, expected_formatted_message", (
    ([AnsibleError('a')], 'a', None),
    ([AnsibleError('a'), AnsibleError('b')], 'a: b', None),
    ([AnsibleError('a: b'), AnsibleError('b')], 'a: b', None),
    ([Exception('a')], 'a', None),
    ([Exception('a'), Exception('b')], 'a: b', None),
    ([Exception('a: b'), Exception('b')], 'a: b', None),
    ([AnsibleError('a'), Exception('b')], 'a: b', None),
    ([AnsibleError('a: b'), Exception('b')], 'a: b', None),
    ([Exception('a'), AnsibleError('b')], 'a: b', None),
    ([Exception('a: b'), AnsibleError('b')], 'a: b', None),
    ([AnsibleError('a'), AnsibleError('b'), Exception('c'), Exception('d')], 'a: b: c: d', None),
    ([AnsibleError('a: b: c: d'), AnsibleError('b'), Exception('c: d'), Exception('d')], 'a: b: c: d', None),

    # collapsing help_text

    (
        [
            AnsibleError('a', help_text='one'),
            Exception('b'),
        ],
        'a: b',
        (
            'a: b one'
        ),
    ),
    (
        [
            AnsibleError('a', help_text='one'),
            AnsibleError('b', help_text='one'),
        ],
        'a: b',
        (
            'a: b one'
        ),
    ),
    (
        [
            AnsibleError('a', help_text='one'),
            Exception('b'),
            AnsibleError('c', help_text='one'),
        ],
        'a: b: c',
        (
            'a: b: c one'
        ),
    ),
    (
        [
            AnsibleError('a', help_text='one'),
            AnsibleError('b', help_text='two'),
        ],
        'a: b',
        (
            'a: b\n\n'
            'a one\n\n'
            '<<< caused by >>>\n\n'
            'b two'
        ),
    ),
    (
        [
            AnsibleError('a'),
            AnsibleError('b', help_text='one'),
        ],
        'a: b',
        (
            'a: b\n\n'
            'a\n\n'
            '<<< caused by >>>\n\n'
            'b one'
        ),
    ),

    # collapsing origin

    (
        [
            AnsibleError('a', obj=_origin_x.tag('x'), orig_exc=Exception('ignored')),
            Exception('b'),
        ],
        'a: b',
        (
            'a: b\n'
            'Origin: /x'
        ),
    ),
    (
        [
            AnsibleError('a', obj=_origin_x.tag('x'), orig_exc=_shared_cause),
            _shared_cause,
        ],
        'a: shared cause',
        (
            'a: shared cause\n'
            'Origin: /x'
        ),
    ),
    (
        [
            # same as above, but exercises the old `orig_exc` path that displays a warning
            AnsibleError('a', obj=_origin_x.tag('x'), orig_exc=Exception('b')),
        ],
        'a: b',
        (
            'a: b\n'
            'Origin: /x'
        ),
    ),
    (
        [
            AnsibleError('a', obj=_origin_x.tag('x')),
            AnsibleError('b', obj=_origin_x.tag('x')),
        ],
        'a: b',
        (
            'a: b\n'
            'Origin: /x'
        ),
    ),
    (
        [
            AnsibleError('a', obj=_origin_x.tag('x')),
            Exception('b'),
            AnsibleError('c', obj=_origin_x.tag('x')),
        ],
        'a: b: c',
        (
            'a: b: c\n'
            'Origin: /x'
        ),
    ),
    (
        [
            AnsibleError('a', obj=_origin_x.tag('x')),
            AnsibleError('b', obj=_origin_y.tag('x')),
        ],
        'a: b',
        (
            'a: b\n\n'
            'a\n'
            'Origin: /x\n\n'
            '<<< caused by >>>\n\n'
            'b\n'
            'Origin: /y'
        ),
    ),
    (
        [
            AnsibleError('a'),
            AnsibleError('b', obj=_origin_y.tag('x')),
        ],
        'a: b',
        (
            'a: b\n\n'
            'a\n\n'
            '<<< caused by >>>\n\n'
            'b\n'
            'Origin: /y'
        ),
    ),
), ids=str)
def test_error_messages(exceptions: list[BaseException], expected_message_chain: str, expected_formatted_message: str | None) -> None:
    with pytest.raises(Exception) as error:
        raise_exceptions(exceptions)

    event = _error_factory.ControllerEventFactory.from_exception(error.value, False)

    message_chain = format_exception_message(error.value)
    formatted_message = _display_utils.format_message(_messages.ErrorSummary(event=event), False)

    assert message_chain == expected_message_chain
    assert formatted_message.strip() == (expected_formatted_message or expected_message_chain)



"""CloudStack plugin for integration tests."""

from __future__ import annotations

import json
import configparser
import os
import urllib.parse
import typing as t

from ....util import (
    ApplicationError,
    display,
)

from ....config import (
    IntegrationConfig,
)

from ....docker_util import (
    docker_exec,
)

from ....containers import (
    run_support_container,
    wait_for_file,
)

from . import (
    CloudEnvironment,
    CloudEnvironmentConfig,
    CloudProvider,
)


class CsCloudProvider(CloudProvider):
    """CloudStack cloud provider plugin. Sets up cloud resources before delegation."""

    def __init__(self, args: IntegrationConfig) -> None:
        super().__init__(args)

        self.image = os.environ.get('ANSIBLE_CLOUDSTACK_CONTAINER', 'quay.io/ansible/cloudstack-test-container:1.7.0')
        self.host = ''
        self.port = 0

        self.uses_docker = True
        self.uses_config = True

    def setup(self) -> None:
        """Setup the cloud resource before delegation and register a cleanup callback."""
        super().setup()

        if self._use_static_config():
            self._setup_static()
        else:
            self._setup_dynamic()

    def _setup_static(self) -> None:
        """Configure CloudStack tests for use with static configuration."""
        parser = configparser.ConfigParser()
        parser.read(self.config_static_path)

        endpoint = parser.get('cloudstack', 'endpoint')

        parts = urllib.parse.urlparse(endpoint)

        self.host = parts.hostname

        if not self.host:
            raise ApplicationError('Could not determine host from endpoint: %s' % endpoint)

        if parts.port:
            self.port = parts.port
        elif parts.scheme == 'http':
            self.port = 80
        elif parts.scheme == 'https':
            self.port = 443
        else:
            raise ApplicationError('Could not determine port from endpoint: %s' % endpoint)

        display.info('Read cs host "%s" and port %d from config: %s' % (self.host, self.port, self.config_static_path), verbosity=1)

    def _setup_dynamic(self) -> None:
        """Create a CloudStack simulator using docker."""
        config = self._read_config_template()

        self.port = 8888

        ports = [
            self.port,
        ]

        descriptor = run_support_container(
            self.args,
            self.platform,
            self.image,
            'cloudstack-sim',
            ports,
        )

        if not descriptor:
            return

        # apply work-around for OverlayFS issue
        # https://github.com/docker/for-linux/issues/72#issuecomment-319904698
        docker_exec(self.args, descriptor.name, ['find', '/var/lib/mysql', '-type', 'f', '-exec', 'touch', '{}', ';'], capture=True)

        if self.args.explain:
            values = dict(
                HOST=self.host,
                PORT=str(self.port),
            )
        else:
            credentials = self._get_credentials(descriptor.name)

            values = dict(
                HOST=descriptor.name,
                PORT=str(self.port),
                KEY=credentials['apikey'],
                SECRET=credentials['secretkey'],
            )

            display.sensitive.add(values['SECRET'])

        config = self._populate_config_template(config, values)

        self._write_config(config)

    def _get_credentials(self, container_name: str) -> dict[str, t.Any]:
        """Wait for the CloudStack simulator to return credentials."""

        def check(value) -> bool:
            """Return True if the given configuration is valid JSON, otherwise return False."""
            # noinspection PyBroadException
            try:
                json.loads(value)
            except Exception:  # pylint: disable=broad-except
                return False  # sometimes the file exists but is not yet valid JSON

            return True

        stdout = wait_for_file(self.args, container_name, '/var/www/html/admin.json', sleep=10, tries=30, check=check)

        return json.loads(stdout)


class CsCloudEnvironment(CloudEnvironment):
    """CloudStack cloud environment plugin. Updates integration test environment after delegation."""

    def get_environment_config(self) -> CloudEnvironmentConfig:
        """Return environment configuration for use in the test environment after delegation."""
        parser = configparser.ConfigParser()
        parser.read(self.config_path)

        config = dict(parser.items('default'))

        env_vars = dict(
            CLOUDSTACK_ENDPOINT=config['endpoint'],
            CLOUDSTACK_KEY=config['key'],
            CLOUDSTACK_SECRET=config['secret'],
            CLOUDSTACK_TIMEOUT=config['timeout'],
        )

        display.sensitive.add(env_vars['CLOUDSTACK_SECRET'])

        ansible_vars = dict(
            cs_resource_prefix=self.resource_prefix,
        )

        return CloudEnvironmentConfig(
            env_vars=env_vars,
            ansible_vars=ansible_vars,
        )



# -*- coding: utf-8 -*-
# Copyright: (c) 2020 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import pytest

from ansible.utils.vars import isidentifier


# Originally posted at: http://stackoverflow.com/a/29586366


@pytest.mark.parametrize(
    "identifier", [
        "foo", "foo1_23",
    ]
)
def test_valid_identifier(identifier):
    assert isidentifier(identifier)


@pytest.mark.parametrize(
    "identifier", [
        "pass", "foo ", " foo", "1234", "1234abc", "", "   ", "foo bar", "no-dashed-names-for-you",
        "True", "False", "None"
    ]
)
def test_invalid_identifier(identifier):
    assert not isidentifier(identifier)


def test_non_ascii():
    """In Python 3 non-ascii characters are allowed as opposed to Python 2. The
    isidentifier method ensures that those are treated as keywords on both
    Python 2 and 3.
    """
    assert not isidentifier("křížek")
